import { StatorContext, StatorObjectBase } from "./statorobject";
import { getMetaApi, getServerEndpoint } from "./ipcbindings";

import { blake2b } from '@noble/hashes/blake2b';
import { bytesToHex } from '@noble/hashes/utils';
import { getNagStator } from "./components/nagscreen";

const ActivationFileName = "activation.json";

//numbers mirrored in activation.h on the C++ side
// and in the website's backend code
//Should there be a separate status for "pending"? As in, we're still *loading*
// whereas *pending" is for delays in talking to the online servers?
enum ActivationStatus
{
    Unlicensed = 0,
    Subscription = 1,
    Perpetual = 2,
    TamperTrip = 3,
    Pending = 4,
    Refunded = 5,
    ChargedBacked = 6,
    InvalidKey = 7,
    ExpiredSub = 8,
    ActivationError = 9,
}

async function getActivationFilePath() {
    return await getMetaApi().joinPath(await getMetaApi().getConfigPath(), ActivationFileName);
}

class Activation extends StatorObjectBase
{
    status = ActivationStatus.Pending;
    licenseKey = "";
    purchaserEmail = "sales@enesda.com"; 
    updatesUntil: number = -3804840000000;
    activationHash: string = "";

    constructor(appContext: StatorContext) {
        super(appContext);
    }

    async save() {
        const data = {
            status: this.status,
            licenseKey: this.licenseKey,
            purchaserEmail: this.purchaserEmail,
            "updatesUntil": this.updatesUntil,
            "activationHash": this.activationHash
        };

        const payload = JSON.stringify(data);
        console.log("payload: " + payload);
        const path = await getActivationFilePath();
        const tempPath = path + ".temp";
        await getMetaApi().writeFile(tempPath, payload, "utf8");
        await getMetaApi().renamePath(tempPath, path);
    }

    calculateActivationHash() {
        const secret = "uwu pwease dwont bweak mwe OWO I hwave bwiws two pway >.<";
        //nodeId in here is kind of a bad idea... for one is it even loaded by this point? And for another...
        // uh... you should be able to change it without borking everything lol
        //It would also encourage people to reuse the machineconfig when they shouldn't because
        // it's the only way to preserve an activation file...
        //This is why you shouldn't second guess things you figured out already...
        const str = `${this.status}${this.licenseKey}${this.purchaserEmail}${this.updatesUntil}${secret}`;
        return `${bytesToHex(blake2b(str))}`;
    }

    async initializeActivation() {
        const readed = await getMetaApi().readUtf8File(await getActivationFilePath());
        if(readed.bytesRead) {
            this.slurp(JSON.parse(readed.data), false);

            const computedHash = this.calculateActivationHash();

            if (computedHash != this.activationHash) {
                console.log("computed hash: " + computedHash);
                console.log("stored hash: " + this.activationHash);
                console.log("tamper tripped! in Initialize!");
                this.status = ActivationStatus.TamperTrip;

                getNagStator(this.getContext()).forceShow();
            }

            //Retry on restart
            if (this.status == ActivationStatus.Pending 
                || this.status == ActivationStatus.ActivationError
                || this.status == ActivationStatus.Subscription) {

                console.log("Phoning home to check license key...")
                await this.activateWithKey(this.licenseKey);
            }
        } else {
            //default values will suffice save for the status
            this.status = ActivationStatus.Unlicensed;
        }
        this.publishChanges();
    }


    private reset()
    {
        this.status = ActivationStatus.Pending;
        //this.licenseKey //we don't reset this! It's ANNOYING
        this.purchaserEmail = "";
        this.updatesUntil = 0;
    }

    shouldNag()
    {
        switch(this.status)
        {
            case ActivationStatus.Pending:
            case ActivationStatus.Subscription:
            case ActivationStatus.Perpetual:
            case ActivationStatus.ActivationError: //This one is false cause it's an error on the backend.
                return false;

            case ActivationStatus.Unlicensed:
            case ActivationStatus.TamperTrip:
            case ActivationStatus.Refunded:
            case ActivationStatus.ChargedBacked:
            case ActivationStatus.InvalidKey:
            case ActivationStatus.ExpiredSub:
                return true;
        }
    }

    private slurp(payload: any, rehash: boolean) {
        try
        {
            this.status = Number(payload["status"]);

            const kickedBackKEy = String(payload["licenseKey"]);
            if (kickedBackKEy)//Cause the server doesn't send bad keys back and it's annoying
                this.licenseKey = kickedBackKEy;

            this.purchaserEmail = String(payload["purchaserEmail"]);
            this.updatesUntil = Number(payload["updatesUntil"]);

            if(rehash) {
                this.activationHash = this.calculateActivationHash();
                this.save().then();
            } else {
                this.activationHash = String(payload["activationHash"]);
            }
        }
        catch(e: any)
        {
            console.warn("Failed to activate with payload");
            console.log(e);
            console.log(payload);
            this.reset();
        }

        this.publishChanges();
    }

    async activateWithKey(key: string)
    {
        //If it a subscription that is already active, and the key is the same, then we're just
        // double checking, otherwise we clear the activation status
        if(this.status != ActivationStatus.Subscription || this.licenseKey != key) {
            this.reset();
            this.licenseKey = key;

            if (!this.licenseKey)
                this.status = ActivationStatus.Unlicensed;

            await this.save();
            this.publishChanges();
        }

        if(!this.licenseKey)
            return;

        const endpoint = (await getServerEndpoint()) + "/api/binder/v1/activate/" + key;

        console.log("Phoning home: requesting activation information");
        const request = new Request(endpoint, { method: "GET" });

        const data = await(await fetch(request)).json();
        console.log(data);

        if (!data)
        {
            console.log("No response from activation servers");
            //Not much we can do about that!
            //But the thing is pending by default so that's fine w/e we don't wanna nag if we don't
            // have our shit together
            return;
        }

        this.slurp(data, true);
    }
}

const ActivationStatorKey = "ActivationStator";

function useWatchActivationStator(appContext: StatorContext)
{
    return appContext.useWatchStatorLegacy(ActivationStatorKey, new Activation(appContext)) as Activation;
}

function getActivationStator(appContext: StatorContext)
{
    return appContext.getStatorLegacy(ActivationStatorKey, new Activation(appContext)) as Activation;
}



export { Activation, useWatchActivationStator, getActivationStator, ActivationStatus }