import React, {useState, useEffect, useRef } from 'react'
import {ObjectGallery, rerenderThresholdFactor} from "./objectgallery";
import { BinderConnectionId, TagId, } from '../backend'
import './tagPage.css'
import {ObjectViewer} from "./objectviewer";
import { StatorContext } from '../statorobject';
import { parseTagName } from '../tagnamefunction';
import { BinderContext, ContextStatus } from './bindercontext';
import { getBinderListStator } from './openbinders';
import { useWatchObservableValue } from '../observermanager';

interface TagPageProps {
    binderContext: BinderContext;
    binderId: BinderConnectionId;
    doRender: boolean;
    appContext: StatorContext;
}

interface TagPageHeaderProps
{
    context: BinderContext;
    myTagId: TagId;
    binderId: BinderConnectionId;
    borkRef: React.MutableRefObject<HTMLDivElement | null>;
    selectedTags: Set<TagId>;
}

enum TagPageLoadState {
    loading,
    loaded,
    syncing,
};

interface TagPageSearchBarProps {
    context: BinderContext;
    query: string;
}

function TagPageSearchBar(props: TagPageSearchBarProps) {
    const [value, setValue] = useState<string>(props.query);

    const ref = useRef<HTMLInputElement>(null);

    const inputHandler = (e: React.ChangeEvent<HTMLInputElement>) => {
        setValue(e.target.value);
    }

    const blurHandler = () => {
        const currentSearchValue = props.context.searchQueryString.get();

        if(!value) {
            //We don't want to clear the search page like that
            if(currentSearchValue) {
                setValue(currentSearchValue);
            }
            return;
        }

        props.context.setSearchQuery(value);
    }

    const submit = (e: any) => {
        e.preventDefault();
        ref.current?.blur();
    }

    return (
        <form onSubmit={submit}>
            <div className='tag-page-search-bar generic-horizontal-input-cont'>
                {
                    //        <h3>Tag Search</h3>
                }
                <input className="tag-page-search-bar-input"

                    ref={ref}
                    onChange={inputHandler}
                    onBlur={blurHandler}
                    value={value}
                    placeholder={"Tag Search"}
                ></input>

                <input value={"Search"} className='generic-horizontal-input-button' type="submit" />

            </div>
        </form>
    );
}

function TagPageTitle(props: TagPageHeaderProps) {
    const [myTagName, setMyTagName] = useState<string>("");
    const [isBeingEdited, setIsBeingEdited] = useState<boolean>(false);

    const editAreaRef = useRef<HTMLDivElement>(null);

    useEffect(() => {
        const nameChangeObsId = props.context.db.registerTagNameChangeObserver(props.myTagId, (name: string) => {
            setMyTagName(name);
        });

        setMyTagName(props.context.db.getTagName(props.myTagId));

        return () => {
            props.context.db.removeTagNameChangeObserver(props.myTagId, nameChangeObsId);
        }
    }, [props.myTagId]);

    useEffect(() => {

    }, []);

    const submitNameChange = async (newName: string) => {
        //This was to check to make sure I wasn't being bitten by ref comparison or something
        //console.log("Name: " + myTagName + " == " + newName + "? ")
        //console.log(myTagName == newName)


        if(newName == myTagName)
            return;

        const succ = await props.context.db.renameTagEndpoint(props.myTagId, newName);

        if(!succ && editAreaRef.current) {
            editAreaRef.current.textContent = myTagName;
        }
    }

    const titleFocusLossHandler = (e: React.FocusEvent) =>
    {
        const content = editAreaRef.current?.textContent;


        //console.log(content);
        submitNameChange("" + content);
        window.getSelection()?.removeAllRanges();
        setIsBeingEdited(false);
    }

    const titleEnterHandler = (e: React.KeyboardEvent) => {
        if (e.key == "Enter")
        {
            e.preventDefault();
;
            //trigger titlefocuslosshandler

            if(editAreaRef == null)
                return;

            if(editAreaRef.current == null)
                return;

            editAreaRef.current.blur();
        }
    }


    const renderedTagName = isBeingEdited ? myTagName : parseTagName(myTagName);
    return (
        <div
            suppressContentEditableWarning={true}
            className="tag-content-view-header-field"
            ref={editAreaRef}
            contentEditable={true}
            onChange={(e: any) => {console.log(e.value)}}
            onKeyDown={titleEnterHandler}
            onBlur={titleFocusLossHandler}
            onFocus={(e: any) => {setIsBeingEdited(true)}}
            spellCheck={isBeingEdited}>
            {renderedTagName}
        </div>
    )
}





//Todo filter from parseTagName when not editing.
function TagPageHeader(props: TagPageHeaderProps) {
    const [searchQuery, setSearchQuery] = useState<string>("");
    const [loadState, setLoadState] = useState<number>(ContextStatus.LoadingTags);

    useWatchObservableValue(props.context.status, (stat: number | undefined) => {
        if(!stat) {
            setLoadState(0);
        } else {
            setLoadState(stat);
        }
    });

    useWatchObservableValue(props.context.searchQueryString, (val: string | undefined) => {
        if (val) {
            setSearchQuery(val);
        } else {
            setSearchQuery("");
        }
    });


    let jsx = <></>;
    let style = {};

    if(loadState == ContextStatus.Ready) {
        if (searchQuery) {
            jsx = <TagPageSearchBar context={props.context} query={searchQuery} />;
        } else if (props.selectedTags.size == 1) {
            jsx = <TagPageTitle {...props} />;
        } else {
            style = { display: "none" };
        }
    }

    //and if selected tags is > 1 then.... Then we have a special search query page :3

 
    return (<div style={style} ref={props.borkRef} className="tag-content-view-header">
        {jsx}
    </div>);
}

enum BottomBarState
{
    GalleryView,
    ObjectViewer,
}

interface ObjectGalleryStatusBarProps
{
    binderContext: BinderContext;
}
function BottomBar(props: ObjectGalleryStatusBarProps) {
    const [queueSize, setQueueSize] = useState<number>(0);
    const [selectedCount, setSelectedCount] = useState<number>(0);

    useEffect(() => {
        const objectCounter = props.binderContext.objectQueue; //kinda hate tracking the whole queue...
        const selectedCounter = props.binderContext.exactSelectedObjectCounter;

        const syncQueueCount = () => {
            const newQeueue = objectCounter.get();

            if(!newQeueue) {
                setQueueSize(0);
            } else {
                setQueueSize(newQeueue.length);
            }
        }

        syncQueueCount();
        const objectCountObsId = objectCounter.register(syncQueueCount);

        const syncSelectedCounter = () => {
            const count = selectedCounter.get();

            if(!count) {
                setSelectedCount(0);
            } else {
                setSelectedCount(count);
            }
        }

        syncSelectedCounter();

        const selectedCountObsId = selectedCounter.register(syncSelectedCounter);

        return () => {
            objectCounter.unregister(objectCountObsId);
            selectedCounter.unregister(selectedCountObsId);
        };
    }, []);

    if(!queueSize)
    {
        return (<div className={"object-gallery-status-bar"}>
            <p> </p>
        </div>);
    }
    else if(selectedCount)
    {
        const percent = Math.round(selectedCount / queueSize * 100);
        return (<div className={"object-gallery-status-bar"}>
            <p>{selectedCount}/{queueSize} objects selected ({percent}%)</p>
        </div>);
    }

    return (<div className={"object-gallery-status-bar"}>
        <p>{queueSize} objects</p>
    </div>);
}

function TagPage(props: TagPageProps) {

    const [inSearchContext, setInSearchContext] = useState<boolean>(false);
    const [selectedTags, setSelectedTags] = useState<Set<TagId>>(new Set<TagId>());
    const [potentiallyInvisibleHeaderHeight, setPotentiallyInvisibleHeaderHeight] = useState<number>(0);
    const [viewAreaHeight, setViewAreaHeight] = useState<number>(0);
    const [nominalScroll, setNominalScroll] = useState<number>(0);

    const scrollableRef = useRef<HTMLDivElement>(null);
    

    const myRef = useRef<HTMLDivElement>(null);
    const headerRef = useRef<HTMLDivElement>(null);
    const galleryRef = useRef(null);
    //These are separate cause I assumed the ObjectViewer is what captured
    // the arrow keys but of course not. If we're re-rendering this errytime
    // anyway this might as well be the same stator...
    //const singleObjectStator = useWatchObjectViewerOpenStator(props.appContext);

    const context = getBinderListStator(props.appContext).getBinder(props.binderId).context;

    let galleryBaseOffset = 0;

    if (selectedTags.size == 1 && !inSearchContext) {
        galleryBaseOffset = potentiallyInvisibleHeaderHeight;
    }

    useEffect(() => {
        const sync = () => {
            if(context.searchQueryString.get()) {
                setInSearchContext(true);
            } else {
                setInSearchContext(false);
            }
        };
        sync();

        const id = context.searchQueryString.register(sync);

        return () => {
            context.searchQueryString.unregister(id);
        }
    },[]);


    useWatchObservableValue(context.currentOpenObject, (val: any) => {
        if(!myRef.current) {
            return;
        }

        if(val) {
            //myRef.current.blur();
        } else {
            //console.log("focusing tag page");
            myRef.current.focus();
        }
    });

    useEffect(() => {

        //We know the tag selection changes when the queue is different sooo
        //It would be better to have a "global observer" on the observer class for
        // the tagSelectionSentinel but we don't so w/e.
        //What's nice here anyway is that we can also bind these changes in the objectGallery :3
        //Which needs its own copy of the object list for some reason I forget
        //Wait it's just for loading state? Why can't we do that here? It's relevant to
        // the tag page...
        const watcher = props.binderContext.tagSelectionChangeWatcher;
        const tagSelectionChangeWatcher = watcher.register(() => {
            //console.log("tag change");
            setSelectedTags(props.binderContext.getSelectedTags());

            if (scrollableRef.current) {
                scrollableRef.current.scrollTop = 0;
            }

        });

        return function () {
            watcher.unregister(tagSelectionChangeWatcher);
        };
    }, []);

    useEffect(() => {
        if(!headerRef) {
            return;
        }
        if (!headerRef.current)
            return;

        const resizeObserver = new ResizeObserver((e: Array<ResizeObserverEntry>) => {
            if (headerRef.current) {

                //console.log("resized header");
                let height = 0;

                height = headerRef.current.offsetHeight;
                const compedStyle = window.getComputedStyle(headerRef.current)

                height += parseInt(compedStyle.getPropertyValue('margin-top'));
                height += parseInt(compedStyle.getPropertyValue('margin-bottom'));

                setPotentiallyInvisibleHeaderHeight(height);

                //props.doRender isn't updated yet when we need this

                    /*
                const rowCount = style.gridTemplateColumns.split(" ").length;

                context.setRowWidth(rowCount);
                */
            }
        });

        resizeObserver.observe(headerRef.current);

        return function () {
            //capture a damn copy of the lil fucker
            const currentRef = headerRef.current;
            if (!currentRef) {
                console.warn("could not unobserve resize, this is okay if this was a live reload");
                return;
            }

            resizeObserver.unobserve(currentRef);
        }
    }, [headerRef]);

    useEffect(() => {
        if(!scrollableRef?.current) {
            return;
        }

        const resizeObserver = new ResizeObserver((e: Array<ResizeObserverEntry>) => {
            if (scrollableRef.current) {

                if(context.currentOpenObject.get()) {
                    //Otherwise it resizes and forces a reload when you close it...
                    return;
                }

                const newHeight = scrollableRef.current.offsetHeight;
                //console.log("new height: " + newHeight);
                setViewAreaHeight(newHeight);

                //props.doRender isn't updated yet when we need this

                    /*
                const rowCount = style.gridTemplateColumns.split(" ").length;

                context.setRowWidth(rowCount);
                */
            }
        });

        resizeObserver.observe(scrollableRef.current);

        return function () {
            //capture a damn copy of the lil fucker
            const currentRef = scrollableRef.current;
            if (!currentRef) {
                console.warn("could not unobserve resize, this is okay if this was a live reload");
                return;
            }

            resizeObserver.unobserve(currentRef);
        }
    }, [scrollableRef.current, viewAreaHeight]);

    /*
    useEffect(() => {
        if(!myRef.current) {
            return;
        }

        myRef.current.addEventListener("scroll", (ev: Event) => {

        });

        return () => {
            if(!myRef.current) {
                return;
            }

        };
    }, [myRef]);
    */

    const handleScroll = (e: any) => {
        //Alright so if it's within a given threshold it'll change the gallery scroll offset, to avoid
        // doing ANYTHING that could trigger a re-render until absolutely necessary.
        if (scrollableRef.current) {
            let s = galleryBaseOffset;

            const myThreshold = viewAreaHeight * rerenderThresholdFactor;
            const adjustedCurrentScroll = scrollableRef.current.scrollTop - galleryBaseOffset;
            
            if(Math.abs(nominalScroll - adjustedCurrentScroll) >= myThreshold) {
                //console.log("Adjusting nominal scroll");
                setNominalScroll(adjustedCurrentScroll);
            }

            /*
            console.log("Gallery offset btw: " + galleryBaseOffset);
            console.log("Scrolled to: " + scrollableRef.current.scrollTop);
            */
        }
    };

    const selectWithOffset = (i: number) =>
    {
        /*
        const mapper = getMapper(props.binderId)

        if(!mapper) return;

        const selected = mapper.getSelected();

        //console.log(selected);



        if(selected.length == 0)
        {
            singleObjectStator.close();
            return;
        }

        let selectedSingle: BinderObject;

        if(selected.length > 1)
        {
            if(i > 0) //Choose the object on the far right end
                selectedSingle = selected[selected.length - 1];
            else //choose the object on the far left end
                selectedSingle = selected[0];
        }
        else
        {
            selectedSingle = selected[0];
        }

        const newSelection = mapper.handleSelectObjectByIndex(selectedSingle.index + i);
        setSingleObject(newSelection.id);
        */
    }


    function downHandler(e: React.KeyboardEvent<HTMLDivElement>) {
        //console.log("Keydown!");
        if(!props.doRender) //invisible so skip
            return;


        //TODO still check if we have focus...
        if (e.ctrlKey && e.key == "a")
        {
            if(e.shiftKey || e.altKey) //I prefer shift but chrome captures "ctrl shift a"
                props.binderContext.invertObjectSelection();
            else
                props.binderContext.selectAllObjects();

            e.preventDefault();
            e.stopPropagation();

            return;
        }
        
        //This is the only case where we need these keys
        if(!props.binderContext.objectsSelectedOrNotSignal.get())
            return;

        if(!e.repeat) //No repeats...
        {
            if(e.key == "ArrowLeft")
            {
                e.preventDefault();
                e.stopPropagation();
                context.selectPrevious();
            }
            if(e.key == "ArrowRight")
            {
                e.preventDefault();
                e.stopPropagation();
                context.selectNext();
            }
            if(e.key == " ")
            {
                e.preventDefault();
                e.stopPropagation();

                if(!e.ctrlKey)
                    context.selectNext();
                else
                    context.selectPrevious();
            }
            if(e.key == "ArrowUp")
            {
                e.preventDefault();
                e.stopPropagation();
                context.selectPreviousRow();
            }
            if(e.key == "ArrowDown")
            {
                e.preventDefault();
                e.stopPropagation();
                context.selectNextRow();
            }
            if(e.key == "Enter")
            {
                e.preventDefault();
                e.stopPropagation();

                if(!context.currentOpenObject.get()) {
                    context.openViewer(undefined); //should fallback on last select index

                } else {
                    context.closeViewer();
                }
            }
            if(e.key == "Escape" && context.currentOpenObject.get())
            {
                e.preventDefault();
                e.stopPropagation();
            }
        }
    }

    //getRootStator(props.appContext).registerDownHandler(downHandler);

    /*
    useEffect(() =>
    {
        //console.log("Registering down handler");
        document.addEventListener("keydown", downHandler)

        return function ()
        {
            //console.log("unregistering down handler");
            document.removeEventListener("keydown", downHandler)
        }
    }, [downHandler, props.doRender]) //Hate that this registers and unregisters every update...

     */


    //Focus on render so that keyboard events are intuitive
    useEffect(() =>
    {
        if(myRef == null)
            return

        if(myRef.current == null)
            return;

        //typescript going crazy on this w/e
        //trus me bro
        // @ts-ignore
        myRef.current.focus();
    });




    const whitespaceClickHandler = (e: React.MouseEvent) => {
        e.stopPropagation();

        if (e.ctrlKey || e.shiftKey) return;

        props.binderContext.clearObjectSelection();
    };



    let myTagId: string = "";

    if (selectedTags.size == 1) {
        selectedTags.forEach((val: string) => {
            myTagId = val;
        });
    }


    //const style = singleObjectStator.viewerOpen ? {display: "none"} : {};

    //The separations of concerns here sucks

    const mouseDownHandler = (e: any) => {
        if(myRef.current) {
            myRef.current.focus();
        }
    };


    return (
        <div className={"tag-page-container"}
        >
            <div className="tag-page"
                tabIndex={1}
                onKeyDown={downHandler}
                onMouseDown={mouseDownHandler}
                ref={myRef}>

                <div
                    ref={scrollableRef}
                    onScroll={handleScroll}
                    className="tag-content-view"
                    onClick={whitespaceClickHandler}
                >
                    <TagPageHeader
                        context={props.binderContext}
                        myTagId={myTagId}
                        binderId={props.binderId}
                        selectedTags={selectedTags}
                        key={myTagId}
                        borkRef={headerRef} />
                    <ObjectGallery
                        viewAreaHeight={viewAreaHeight}
                        nominalOffset={nominalScroll}
                        theBinderId={props.binderId}
                        tagSelectionCount={selectedTags.size}
                        appContext={props.appContext}
                    />
                </div>


                <BottomBar binderContext={props.binderContext} />
            </div>
            <ObjectViewer
                theBinderId={props.binderId}
                appContext={props.appContext}
            />
        </div>
    );
}

export { TagPage, TagPageLoadState }