import React from 'react';
import {StatorContext, StatorObjectBase} from "../statorobject";
import './uploadcontroller.css';

enum UploadState {
    Preparing,
    InProgress,
    Error,
    Finished,
    Cancelled,
}

export class UploadControllerStator extends StatorObjectBase
{
    private show: boolean = false;
    private totalCount: number = 0;
    private succeededCount: number =0;
    private failedCount: number =0;
    private message: string = "";
    private state = UploadState.Preparing;
    private error: string = "";
    private fileName: string = "";


    nextFileName(name: string) {
        //increment current count, set name
        this.fileName = name;
        this.state = UploadState.InProgress;
        this.publishChanges();
    }

    setPreparing() {
        this.succeededCount = 0;
        this.failedCount = 0;
        this.fileName = "";
        this.totalCount = 0;
        this.show = true;
        this.state = UploadState.Preparing;
        this.message = "";
        this.error = "";
        this.publishChanges();
    }

    setCount(ct: number) {
        //Maybe all this should be in a struct destroyed on completion?
        this.totalCount = ct;
        this.publishChanges();
    }

    getFileName()
    {
        return this.fileName;
    }

    getMessage()
    {
        return this.message;
    }

    getTotalCount()
    {
        return this.totalCount;
    }

    getSuccessCount() {
        return this.succeededCount;
    }

    getFailCount() {
        return this.failedCount;
    }

    succ() {
        ++this.succeededCount;
        this.publishChanges();
    }

    fail(msg: string)
    {
        ++this.failedCount;
        this.publishChanges();
    }

    errorOut(msg: string) {
        this.message = msg;
        this.state = UploadState.Error;
        this.publishChanges();
    }

    finish() {
        this.state = UploadState.Finished;
        this.publishChanges();
    }

    checkShow() {
        return this.show;
    }

    checkIsCancelled() {
        return this.state == UploadState.Cancelled;
    }

    close()
    {
        this.show = false;
        this.publishChanges();
    }

    cancel() {
        this.state = UploadState.Cancelled;
        this.publishChanges();
    }

    getState() {
        return this.state;
    }

}

const UploadControllerStatorKey = "UploadControllerStator";

function useWatchUploadControllerStator(appContext: StatorContext): UploadControllerStator
{
    return appContext.useWatchStatorLegacy(UploadControllerStatorKey, new UploadControllerStator(appContext)) as UploadControllerStator;
}

function getUploadControllerStator(appContext: StatorContext): UploadControllerStator
{
    return appContext.getStatorLegacy(UploadControllerStatorKey, new UploadControllerStator(appContext)) as UploadControllerStator;
}

export interface UploadControllerProps
{
    appContext: StatorContext;
}

function UploadController(props: UploadControllerProps)
{
    const myStator = useWatchUploadControllerStator(props.appContext);

    if(!myStator.checkShow())
        return <></>;

    function closer(e: any)
    {
        e.preventDefault();
        if(myStator.getState() == UploadState.InProgress)
            myStator.cancel();
        else if(myStator.getState() == UploadState.Preparing) {
            /*noop*/
        } else {
            myStator.close();
        }
    }

    const closeButtonValue = myStator.getState() != UploadState.InProgress ? "Close" : "Cancel";

    const closeButtonJsx = myStator.getState() == UploadState.Preparing ? (<></>) : 
        (

            <form onSubmit={closer}>
                <input type="submit" className="generic-overlay-form-button" value={closeButtonValue} />
            </form>
        )

    let title = "";
    let content = "";
    let progressString = `${myStator.getSuccessCount()}/${myStator.getTotalCount()}`;
    switch(myStator.getState()) {
        case UploadState.Cancelled:
            title = "Cancelled";
            content = progressString;
            break;
        case UploadState.Finished:
            title = "Done!";
            const failCount = myStator.getFailCount();
            if(failCount != 0) {
                content = `Added ${progressString} files, failed to add ${failCount} files :< sorry`;
            } else {
                content = `Added ${progressString} files`;
            }
            break;
        case UploadState.Error:
            title = "Oops!"
            content = myStator.getMessage();
            break;
        case UploadState.InProgress:
            title = "Adding...";
            content = `${progressString}: ${myStator.getFileName()}`;
            break;
        case UploadState.Preparing:
            title = "Scanning...";
            break;
    }

    /*
    const contensist = myStator.checkIsFinished() ? (<p>Added {myStator.getCurrentCount()} files.</p>) :
        (<p>{myStator.getCurrentCount()}/{myStator.getTotalCount()}: {myStator.getFileName()}</p>);
        */

    return (
        <div className={"upload-controller-background"} draggable={false}>
            <div className={"center-overlay"}>
                <div className={"generic-overlay-content-container"}>
                    <div className={"generic-overlay-content-container"}>
                        <div className={"generic-overlay-inner"}>
                            <h1>{title}</h1>
                            <p>{content}</p>
                            {closeButtonJsx}
                        </div>
                    </div>
                </div>
            </div>
        </div>
    );
}

export {UploadController, getUploadControllerStator}