import {
    assertGoodWriteRecord,
    constructActiveProperty,
    createRecord,
    digestRecordIntoProperty, readPropertyString,
} from "./writerecord.js";


function constructCache() {
    const obj = {};

    obj.tables = {};
    obj.indexes = {}; //Because it's the plural of an index as in a list, not as in indices
    obj.lastReadByMachine = {}; //machineId -> last read value

    return JSON.stringify(obj);
}

const keyDelimiter = ":";

//IMPRTANTNLY: this can be called on the result and it will be the same,
// this allows us to use the composite key that is made internally,
// with functions exposed to the external api
function makeCompositeKey(...keys) {
    let result = "";

    for(const k of keys) {
        if(k.includes(keyDelimiter)) {
            throw new Error(`Nucleus2: Invalid key, contains reserved character: ${keyDelimiter}`);
        }

        if(result === "")
            result = k;
        else
            result = `${result}${keyDelimiter}${k}`
    }

    return result;
}

//key1 key2 key3
//up to: key3 is a value key
//up to: key2 is an index key
//up to: key1 is not much (majors?)

//example 2:

//key1 key2 key3 key4
//up to: key4 is a value key
//up to: key3 is an index key
//up to: key2 is not much (majors?)

//example 3:

//key1 key2
//up to: key2 is a value key
//up to: key1 is an index key

//key1
//up to: key1 is a value key
//there is no index key

function fractureKey(obsKey) {
    //I kind of hate this, but what would a clean alternative even be?
    // either this or passing around all the component keys and compositing
    // them all the time which would be even more often than this...
    const splitKey = obsKey.split(keyDelimiter);

    if(splitKey.length === 0)
        throw new Error("Nucleus2: error, composite key too short to split");

    return splitKey;
}

//The last key on it's own is called the "minor" key

function popKey(obsKey) {
    const splitKey = fractureKey(obsKey);
    const minorKey = splitKey.pop(); //come on errybody let's do the split key pop

    //makeComposite just returns an empty key with an empty input
    return [makeCompositeKey(...splitKey), minorKey];
}

class Nucleus2 {
    #nodeId = "";
    #cacheable = {}; // all the data we need to dump to disk and read back in to restore committed db state
    #cacheFlushLock = false;

    #syncLock = false;

    #queued = {};

    #flushAttemptedDuringFlush = false;
    #commitLock = false;
    #pendingCommit = {};

    #propertyObservers = {}; // observerKey -> {index: callback}
    #indexObservers = {};
    #majorObservers = {};
    #lastObserverId = 0;

    #outdatedPropertyObservers = {};
    #outdatedIndexObservers = {};


    //Is it this simply simple?
    #logReader;
    #logWriter;

    //#cacheReader; //why would we need to read our own cache more than once?
    #cacheWriter;

    //Logreader is expected to read all logs of each node, returned in a list as such:
    //{machine: {logs: [...], readTo: number}, ...}
    constructor(nodeId, logReader, logWriter, cacheReader, cacheWriter) {
        this.#nodeId = nodeId;
        this.#logReader = logReader;
        this.#logWriter = logWriter;
        this.#cacheWriter = cacheWriter;


        //Initialize cache, expected to return a newly constructed cache object
        // if nothing is saved already.
        this.#cacheable = JSON.parse(cacheReader());
    }

    #valueChanged(obsKey) {
        const val = this.#getPropertyString(obsKey);
        const [indexKey, minorKey] = popKey(obsKey);

        let indexChanged = false;

        if(val !== undefined) {
            if(!(indexKey in this.#cacheable.indexes)) {
                this.#cacheable.indexes[indexKey] = {};
            }

            this.#cacheable.indexes[indexKey][minorKey] = null;

            indexChanged = true;

        } else if(indexKey in this.#cacheable.indexes) { //val is undefined and the index exists

            //remove it
            if(minorKey in this.#cacheable.indexes[indexKey]) {
                delete this.#cacheable.indexes[indexKey][minorKey];
                indexChanged = true;
            }

            //Delete index too, if it's now empty
            if(Object.keys(this.#cacheable.indexes[indexKey]).length === 0) {
                delete this.#cacheable.indexes[indexKey];
                //indexChanged should already be set here
            }
        }


        this.#outdatedPropertyObservers[obsKey] = null;

        if(indexChanged) {
            this.#outdatedIndexObservers[indexKey] = null;
        }
    }

    #alertDeferredObservers() {
        for (const obsKey in this.#outdatedPropertyObservers) {
            const val = this.#getPropertyString(obsKey);
            let valToPropagate = undefined;
            if (val !== undefined)
                valToPropagate = JSON.parse(val);

            for (const observerId in this.#propertyObservers[obsKey]) {
                this.#propertyObservers[obsKey][observerId](valToPropagate, () => this.removePropertyObserver(...fractureKey(obsKey), observerId));
            }

            const observedPath = fractureKey(obsKey);
            let parentKey = obsKey;



            while(true) {
                [parentKey, ] = popKey(parentKey);
                //console.log("parent key: " + parentKey);

                if(!parentKey) break;


                if (!(parentKey in this.#majorObservers)) {
                    continue;
                }

                
                for(const observerId in this.#majorObservers[parentKey]) {
                    const fracturedKey = fractureKey(parentKey);
                    this.#majorObservers[parentKey][observerId](observedPath, valToPropagate, () => this.removeMajorObserver(...fracturedKey, observerId));
                }
            }
        }


        for (const indexKey in this.#outdatedIndexObservers) {
            for(const observerId in this.#indexObservers[indexKey]) {
                this.#indexObservers[indexKey][observerId](() => this.removeIndexObserver(...fractureKey(indexKey), observerId));
            }
        }


        this.#outdatedPropertyObservers = {};
        this.#outdatedIndexObservers = {};
    }

    async flushCache() {
        if(this.#cacheFlushLock) {
            throw new Error("Nucleus2: Can't flush cache with flush in progress");
        }
        this.#cacheFlushLock = true;
        await this.#cacheWriter(JSON.stringify(this.#cacheable));
        this.#cacheFlushLock = false;
    }

    #getPropertyString(obsKey) {
        if(obsKey in this.#queued) {
            const rec = this.#queued[obsKey];
            if(rec.recordType === "delete") {
                return undefined;
            }

            return rec.value;
        }

        if(obsKey in this.#pendingCommit) {
            const rec = this.#pendingCommit[obsKey];
            if(rec.recordType === "delete") {
                return undefined;
            }

            return rec.value;
        }

        if(!(obsKey in this.#cacheable.tables))
            return undefined;

        //Read property value/string also returns undefined if the value is deleted :3
        return readPropertyString(this.#cacheable.tables[obsKey]);
    }

    get(...args) {
        if(args.length === 0)
            throw new Error("Nucleus2: error must provide arguments to get");

        const val = this.#getPropertyString(makeCompositeKey(...args));

        if(val === undefined)
            return undefined;

        return JSON.parse(val);
    }

    checkCommitLock() {
        return this.#commitLock;
    }

    checkSyncLock() {
        return this.#syncLock;
    }

    set(...args) {
        if(args.length === 0)
            throw new Error("Nucleus2: Missing arguments to set");

        if(args.length === 1)
            throw new Error("Nucleus2: Missing value/key in set call");

        const value = args.pop();

        if(value === undefined) {
            throw new Error("Nucleus2: cannot set value to undefined, use null or delete it");
        }

        const obsKey = makeCompositeKey(...args);

        if(this.#getPropertyString(obsKey) === JSON.stringify(value)) //noop
            return;

        if(!(obsKey in this.#cacheable.tables)) {
            this.#cacheable.tables[obsKey] = constructActiveProperty(this.#nodeId, obsKey);
        }

        const prop = this.#cacheable.tables[obsKey];

        this.#queued[obsKey] = createRecord(prop, "write", value);
        this.#valueChanged(obsKey);
        this.#alertDeferredObservers();
    }

    del(...args) {
        if(args.length === 0)
            throw new Error("Nucleus2: error, must pass args to delete");

        const obsKey = makeCompositeKey(...args);

        if(this.#getPropertyString(obsKey) === undefined) //doesn't exist or was already deleted
            return;

        if(!(obsKey in this.#cacheable.tables)) //definitely doesn't exist
            return;

        const prop = this.#cacheable.tables[obsKey];

        //It has to be made against the property and not the record in queue.
        // and we don't want to digest that one until it's committed, so we just accept clashing
        // machine indices for overwrites (they'll resolve safely with timestamps, since it's
        // the same process)
        this.#queued[obsKey] = createRecord(prop, "delete");
        this.#valueChanged(obsKey);
        this.#alertDeferredObservers();
    }


    async sync() {
        //If we're committing, it's fine, order of digestRecordInto property is completely arbitrary :)
        if(this.#syncLock) {
            throw new Error("Nucleus2: Error, cannot sync with sync in progress!");
        }

        this.#syncLock = true;
        const alertables = {};
        let failedToFullyRead = false;

        try {
            //Do the things

            //Logreader is expected to read all logs of each node, returned in a list as such:
            //{machine: {logs: [...], readTo: number}, ...}

            const logs = await this.#logReader(this.#cacheable.lastReadByMachine);

            for(const machineId in logs) {
                const logBlock = logs[machineId];

                if(!("logs" in logBlock)) {
                    console.log("Nucleus2: log block missing logs field");
                    continue;
                }

                if(!("readTo" in logBlock)) {
                    console.log("Nucleus2: log block missing readTo field");
                    continue;
                }

                for(const record of logBlock.logs) {
                    assertGoodWriteRecord(record);

                    const obsKey = record.key;

                    //Store the current value the first time we encounter it
                    if(!(obsKey in alertables)) {
                        alertables[obsKey] = this.#getPropertyString(obsKey);
                    }

                    if(!(obsKey in this.#cacheable.tables)) {
                        this.#cacheable.tables[obsKey] = constructActiveProperty(
                            this.#nodeId, obsKey);
                    }

                    digestRecordIntoProperty(this.#cacheable.tables[obsKey], record);
                }

                //Set this after actually reading everything, in case there's an throw above.
                // that way we don't get stuck. This is essentially a commit.
                this.#cacheable.lastReadByMachine[machineId] = logBlock.readTo;
            }
        } catch(e) {
            failedToFullyRead = true;
        }

        this.#syncLock = false;

        //Not really necessary now that we've batched all alerts :3c
        for(const obsKey in alertables) {
            if(alertables[obsKey] !== this.#getPropertyString(obsKey)) {
                this.#valueChanged(obsKey);
            }
        }

        this.#alertDeferredObservers();

        if(failedToFullyRead)
            throw new Error("Nucleus2: Failed to fully read logs, might have read some or none");
    }


    async flushToPersistent() {
        if(this.#commitLock) {
            this.#flushAttemptedDuringFlush = true;
            return;
            //throw new Error("Nucleus2: Error, cannot flush with flush in progress!");
        }

        while(true) {
            const alertables = {};
            let failedToWrite = false;

            this.#commitLock = true;

            this.#pendingCommit = this.#queued;
            this.#queued = {};

            const batchedRecords = [];

            for (const obsKey in this.#pendingCommit) {
                const record = this.#pendingCommit[obsKey];
                batchedRecords.push(JSON.stringify(record));
            }

            try {
                //send it
                await this.#logWriter(batchedRecords);

                for (const obsKey in this.#pendingCommit) {
                    const record = this.#pendingCommit[obsKey];

                    if (!(obsKey in this.#cacheable.tables)) {
                        console.log("Nucleus2: warning, table not found.");
                        //I mean, you have to have a property to create a record, so
                        // this is like... a big deal.
                        continue;
                    }

                    const committedProp = this.#cacheable.tables[obsKey];
                    //We only have to reconcile removal of minor keys from the lists
                    // if it's committed, the listMinors function looks into the
                    // queued and pending to prune the result, if we just remove a
                    // failed record from there, then it's fine.

                    digestRecordIntoProperty(committedProp, record);
                    //assume observers already up to date, because optimistic writes
                }
            } catch (e) {
                failedToWrite = true;
                for (const obsKey in this.#pendingCommit) {
                    const record = this.#pendingCommit[obsKey];

                    if (obsKey in this.#queued) //already obsolete so just roll with it
                        continue;

                    let committedVal;

                    if (!(obsKey in this.#cacheable.tables)) {
                        console.log("Nucleus2: warning: table not found this should never happen");
                        //Continue to alert anyway
                        committedVal = undefined;
                    } else {
                        committedVal = readPropertyString(this.#cacheable.tables[obsKey]);
                    }

                    if (committedVal !== record.value)
                        alertables[obsKey] = true;
                }
            }


            this.#pendingCommit = {}; //wait until after the record is digested to reset pending

            //AFTER pending is cleared
            for (const obsKey in alertables) {
                this.#valueChanged(obsKey);
            }

            this.#alertDeferredObservers();

            this.#commitLock = false;

            if (failedToWrite)
                throw new Error("Nucleus2: failed to write to persistent, db is in recovered state");

            if(this.#flushAttemptedDuringFlush) {
                this.#flushAttemptedDuringFlush = false;
                continue;
            } else {
                break;
            }
        }
    }

    //E.g. table: tagLinks, major: tagId, minor: objId
    //list("tagLinks", "$tagId$") - > list of objects with tag
    listMinors(...args) {

        if(args.length === 0)
            throw new Error("Nucleus2: error, must pass a key to list minors");

        const minorsKey = makeCompositeKey(...args);

        if(!(minorsKey in this.#cacheable.indexes))
            return [];

        return Object.keys(this.#cacheable.indexes[minorsKey]);
    }

    #observerRegistrationHelper(dest, ...args) {
        if(args < 2) {
            throw new Error("Nucleus2: error, invalid number of arguments to register observer");
        }

        const observer = args.pop();
        const key = makeCompositeKey(...args);

        const observerId = ++this.#lastObserverId;

        if(!(key in dest)) {
            dest[key] = {};
        }

        dest[key][observerId] = observer;

        return observerId;
    }

    #observerRemovalHelper(location, ...args) {
        if(args < 2) {
            throw new Error("Nucleus2: error, invalid number of arguments to remove property observer");
        }

        const id = args.pop()
        const key = makeCompositeKey(...args);
        if(!(key in location))
            return;

        if(!(id in location[key]))
            return;

        delete location[key][id];

        if(location[key].length === 0) {
            //Delete the whole observer list if empty
            delete location[key];
        }
    }

    //keys..., observer
    registerPropertyObserver(...args) {
        return this.#observerRegistrationHelper(this.#propertyObservers, ...args);
    }

    removePropertyObserver(...args) {
        this.#observerRemovalHelper(this.#propertyObservers, ...args);
    }

    registerIndexObserver(...args) {
        return this.#observerRegistrationHelper(this.#indexObservers, ...args);
    }

    removeIndexObserver(...args) {
        this.#observerRemovalHelper(this.#indexObservers, ...args);
    }

    registerMajorObserver(...args) {
        return this.#observerRegistrationHelper(this.#majorObservers, ...args);
    }

    removeMajorObserver(...args) {
        this.#observerRemovalHelper(this.#majorObservers, ...args);
    }
}

export default Nucleus2;
export {constructCache}