import { Binder, BinderConnectionId } from "./backend";
import { getMetaApi } from "./ipcbindings";
import {v4 as idGen} from 'uuid';
import { ObservableValue, differenceFilter } from "./observermanager";
import { BinderContext } from "./components/bindercontext";
import { takeLicenseChecksum } from './components/eula';

const LocalConfigFilename = "machinespecificconfig.json";
const BinderIdFilename = "binder-id";

const nodeIdField = "nodeId";
const agreedToEulaFieldName = "versionsOfLicenseAgreedTo";

//And binderId now is just the UUID used to identify the folder, and not simply a "DB*" name
const clusterConField = "clusterConnections"; //[path]

const clusterStageFolder = "Stage";
const clusterStandardThumbnailsFolder = "StandardThumbnails";
const clusterObjectsFolder = "Objects";
const clusterWriteLogsFolder = "WriteLogs";


class LocalNodeConfig {
    readonly connectedBinders = new ObservableValue<Array<BinderConnectionId>>(differenceFilter)

    private static alreadyExists: boolean = false;
    private localConfigPath: string | undefined;
    private data: any = {};
    private saveLock = false;
    private loadLock = false;
    private loaded = false;

    constructor() {
        if(LocalNodeConfig.alreadyExists) {
            throw new Error("Cannot initialize a second copy of the local config");
        }

        this.data[nodeIdField] = idGen();
        this.data[clusterConField] = [];
        this.data[agreedToEulaFieldName] = false;
        this.localConfigPath = undefined;
        LocalNodeConfig.alreadyExists = true;
    }

    private async getPath(): Promise<string> {
        if(this.localConfigPath) {
            return this.localConfigPath;
        }

        const result = await getMetaApi().joinPath(await getMetaApi().getConfigPath(), LocalConfigFilename);

        if(!result) {
            throw new Error("Could not get local config path");
        }

        return result;
    }

    checkLoadInitiated() {
        return this.loadLock;
    }

    checkLoaded() { return this.loaded; }

    async load() {
        if(this.loadLock) {
            throw Error("Cannot initialize second loading of LocalNodeConfig");
        } else {
            this.loadLock = true;
        }

        if(this.saveLock) {
            throw Error("STuck in save lock cannot load");
        }

        const payload = await getMetaApi().readUtf8File(await this.getPath());

        if(payload.bytesRead) {
            try {
                //console.log("about to parse: " + JSON.stringify(payload.data));
                this.data = JSON.parse(payload.data);
                //validate maybe????? or just fail and it'll catch and default initialize...

                //Updat'em :3
                this.connectedBinders.set([...this.data[clusterConField]]);
                console.log("Loaded local config");
                this.loaded = true;
            } catch(e: any) {
                console.warn("Failed to parse localconfig");
            }
        }
    }

    async save() {
        if(this.saveLock) {
            console.warn("Attempt to save during save");
            return;
        }

        if(!this.loadLock) {
            throw Error("Cannot save config before load");
        }

        try {
            this.saveLock = true;
            const path = await this.getPath();
            const tempPath = path + ".temp";
            await getMetaApi().writeFile(tempPath, JSON.stringify(this.data), "utf8");
            await getMetaApi().renamePath(tempPath, path);
        } finally {
            this.saveLock = false;
        }
    }

    private readLicenseAgreementsAsArray() {
        let result = new Array<String>();

        if (this.data[agreedToEulaFieldName])
            result = new Array<String>(...this.data[agreedToEulaFieldName]);

        console.log("Versions agreed: " + JSON.stringify(result));
        return result;
    }

    async checkEulaAgreement(): Promise<boolean> {
        return this.readLicenseAgreementsAsArray().includes(await takeLicenseChecksum());
    }

    async setEulaAgreed() {
        const agreements = this.readLicenseAgreementsAsArray();
        agreements.push(await takeLicenseChecksum());
        this.data[agreedToEulaFieldName] = agreements;
        await this.save();
    }

    getLocalNodeId() {
        return this.data.nodeId;
    }

    getBinderList() {
        return this.data[clusterConField];
    }

    async checkBinderExists(id: BinderConnectionId) {
        if(!this.data[clusterConField].includes(id)) {
            return false;
        }

        try {
            const idFile = (await this.readBinderIdFile(id));

            if(!idFile) {
                return false;
            }
        } catch(e) {
            return false;
        }

        return true;
    }

    private async readBinderIdFile(path: string) {
        let result = (await getMetaApi().readUtf8File(await getMetaApi().joinPath(path, BinderIdFilename))).data;
        if(typeof result == "string") {
            result = result.trim();
        }
        return result ? result : undefined;
    }

    async registerCluster(path: string) {
        if(this.data[clusterConField].includes(path)) {
            return "Binder already registered";
        }
        //just check if the binder-id file exists, if it does we can relatively safely assume it's
        // a binder. We generate a connection ID so that we can connect to manually duplicated binders
        const binderFolderId = await this.readBinderIdFile(path);

        //We didn't impelment the checkExists function...

        if (!binderFolderId) {
            return "Could not connect to binder, missing binder-id file";
        }

        console.log("Adding binder with folder id: " + binderFolderId);

        this.data[clusterConField].push(path);
        this.connectedBinders.set([...this.data[clusterConField]]);

        //Todo: if a save is in progress, unlikely but here me out, then we'll drop this binder
        // inadvertently...
        await this.save();

        return "";


        /*

        const conId = (++this.data[conIdCountField]).toString();

        //why even have this????
        // like??? just use the path??? The idea was to not leak the actual path to the frontend I think
        // but we're waaaaaaaaaaaaaaaaaaaaaaaaaaay past that now.
        // now we'd have to iterate over everything or have a second set for just the connectino paths
        // to prevent duplicates...
        */


        /*
        const binderId = await getMetaApi().readUtf8File(await getMetaApi().joinPath(path, BinderIdFilename));
        binderId.trim();
    
        if(binderId in this.data[clusterConField]) {
            //I mean hypothetically if the id isn't unique this could fail but hecc
            // we don't support tampering with these files...
            //but........
            //.............. you could maybe wanna duplicate it naively like that...
            console.log("A binder with that ID already exists, skipping add");
            return;
        }
        */
    }

    async removecluster(id: BinderConnectionId) {
        const index = this.data[clusterConField].indexOf(id);
        if(index < 0) {
            return "Binder does not exist in localconfig, so cannot be removed";
        }

        this.data[clusterConField].splice(index, 1);
        await this.save();

        this.connectedBinders.set([...this.data[clusterConField]]);
    }

    async loadBinder(id: BinderConnectionId): Promise<Binder> {
        const result = new Binder(new BinderContext(id));

        result.id = id;
        result.name = await getMetaApi().baseNameOfPath(id); //FIXME needs a file system path.whatever call...
        result.exists = await this.checkBinderExists(id);
        result.clusterPath = id;

        return result;
    }

    //-----------------------------------------------Getters/generators for paths
    private async generateSubPathHelper(id: BinderConnectionId, base: string, subpath: string) {
        if(!await this.checkBinderExists(id)) {
            throw new Error("Binder does not exist no path");
        }

        const result = await getMetaApi().joinPath(base, subpath);
        await getMetaApi().createDirectories(result);

        return result;
    }

    async getPathForStage(id: BinderConnectionId) {
        return this.generateSubPathHelper(id, id, clusterStageFolder);
    }

    async getPathForThumbnails(id: BinderConnectionId) {
        return this.generateSubPathHelper(id, id, clusterStandardThumbnailsFolder);
    }

    async getPathForObjects(id: BinderConnectionId) {
        return this.generateSubPathHelper(id, id, clusterObjectsFolder);
    }

    async getWriteLogDirectory(id: BinderConnectionId) {
        return this.generateSubPathHelper(id, id, clusterWriteLogsFolder);
    }

    async getPathForLocalWriteLogs(id: BinderConnectionId) {
        return this.generateSubPathHelper(id, await this.getWriteLogDirectory(id), this.getLocalNodeId());
    }
};

const localNodeConfig = new LocalNodeConfig();

export {localNodeConfig, BinderIdFilename}
