import { blake2b } from '@noble/hashes/blake2b';
import { bytesToHex } from '@noble/hashes/utils';
import { getMetaApi } from './ipcbindings';
import { findEndOfUnboundedList } from './exponentialsearch';

//The number of directories deep to store the actual write log.
//This amounts to 2^4 logs before we go over 256 files per leaf
// directory in the average case
const writeLogPathDepth = 3;
const writeLogSoftLimitBytes = 1024 * 1024 / 2; // 1/2mb :3

interface LogFileData {
    precedingCount: number,
    limitBytes: number,
    records: string[],
}

const example = {
    precedingCount: 12938,
    limitBytes: 1048576, //this way we can change it and still honor the old version
    records: ["etc", "etc", "etc"],
};

function fracturePath(hex: string, depth: number) {
    const result = new Array<string>();
    const requiredChars = depth * 2; //two chars per byte, one byte per level

    if (hex.length < requiredChars)
        throw new Error("Value for path fracturing too short");

    let pos = 0;

    for (let i = 0; i != requiredChars; /*NoO!*/) {
        let level = "";
        level += hex[pos];
        level += hex[pos + 1];

        result.push(level);

        pos += 2;
        i += 2;
    }

    return result;
}

class LogFile {
    private full = false;

    //Because I don't want to create the directories until we're ready to write, okay??
    private dirPath: string;
    private filename: string;

    private index: number;
    private data: LogFileData;

    constructor(dirPath: string, filename: string, index: number) {
        this.dirPath = dirPath;
        this.filename = filename;

        this.index = index;

        this.data = {
            precedingCount: 0,
            limitBytes: writeLogSoftLimitBytes,
            records: [],
        }
    }

    setPreceding(count: number) {
        this.data.precedingCount = count;
    }

    getPreceding() {
        return this.data.precedingCount;
    }

    getCount() {
        return this.data.records.length;
    }

    getIndex(): number {
        return this.index;
    }

    isFull(): boolean {
        return this.full;
    }

    digest(records: string[]) {
        for(const rec of records) {
            const isString = typeof rec == 'string';
            //console.log("Digesting record, is string? " + isString);
            this.data.records.push(rec);
        }
    }

    getRecords() {
        return [...this.data.records];
    }

    async load() {
        const path = await getMetaApi().joinPath(this.dirPath, this.filename);
        //console.log("Loading " + path);
        const payload = await getMetaApi().readUtf8File(path);
        //console.log("Payload size: " + payload.bytesRead);

        if(!payload.bytesRead) {
            console.warn("Payload empty in load logfile, read: " + payload.bytesRead);
            return;
        }


        const data = JSON.parse(payload.data);

        let corrupt = false;

        if(!(typeof data.precedingCount == 'number')) {
            corrupt = true;
        }

        if(!(typeof data.limitBytes == 'number')) {
            corrupt = true;
        }

        if(!(typeof data.records == 'object')) { //because we can't have 'array'...
            corrupt = true;
        }

        if(!corrupt) {
            this.data = data;
            
            if (payload.bytesRead >= this.data.limitBytes) {
                this.full = true;
            }
        } else {
            console.warn("write log is corrupt, will be overwritten")
        }
    }

    async write() {
        //console.log("Writing LogFile");
        const payload = JSON.stringify(this.data);

        await getMetaApi().createDirectories(this.dirPath);
        const tempFilename = this.filename + ".temp";
        const tempPath = await getMetaApi().joinPath(this.dirPath, tempFilename);
        await getMetaApi().writeFile(tempPath, payload, "");

        const commitPath = await getMetaApi().joinPath(this.dirPath, this.filename);
        await getMetaApi().renamePath(tempPath, commitPath);

        if(payload.length >= this.data.limitBytes) {
            this.full = true;
        }
    }
}

async function getLogFilename(clusterPath: string, nodeId: string, index: number): Promise<string> {
    return `${bytesToHex(blake2b(`${nodeId}${index}`))}.json`;
}

async function genFracturedPath(basePath: string, hex: string, depth: number = 3) {
    let path = basePath;

    for (const part of fracturePath(hex, depth)) {
        path = await getMetaApi().joinPath(path, part);
    }

    return path;
}

async function getLogfileDir(clusterPath: string, filename: string): Promise<string> {
    return genFracturedPath(clusterPath, filename, writeLogPathDepth);
}

async function getLogFilePath(clusterPath: string, nodeId: string, index: number): Promise<string> {
    const filename = await getLogFilename(clusterPath, nodeId, index);
    const logFileDir = await getLogfileDir(clusterPath, filename);

    return await getMetaApi().joinPath(logFileDir, filename);
}

async function checkLogFileExists(clusterPath: string, nodeId: string, index: number): Promise<boolean> {
    const path = await getLogFilePath(clusterPath, nodeId, index);
    return await getMetaApi().checkPathExists(path);
}

async function openLogFileForIndex(localLogsPath: string, nodeId: string, index: number) {
    const logFilename = await getLogFilename(localLogsPath, nodeId, index);
    const logDirPath = await getLogfileDir(localLogsPath, logFilename);

    return new LogFile(logDirPath, logFilename, index);
}

async function openLatestLogFile(localLogsPath: string, nodeId: string, startingPoint: number) {
    const predicate = async (index: number) => await checkLogFileExists(localLogsPath, nodeId, index);
    const index = await findEndOfUnboundedList(startingPoint, predicate);
    return openLogFileForIndex(localLogsPath, nodeId, index);
}

export {openLatestLogFile, openLogFileForIndex, genFracturedPath}