import { ObservableCounterMap, ObservableMap, differenceFilter, zeroOrManyFilter } from "./observermanager";

test('Observable map, register and set (difference filter)', () => {
    const key = "kiki";
    const subject = new ObservableMap<string, boolean | undefined>(differenceFilter);

    let count = 0;

    const observer = () => {
        ++count;
    };

    subject.register(key, observer);
    expect(subject.hasWatchersOrValue(key)).toBeTruthy();
    subject.set(key, true);

    expect(count).toBe(1);
});

test('Observable map, register and set, set to undefined, unregister (difference filter)', () => {
    const key = "kiki";
    const subject = new ObservableMap<string, boolean | undefined>(differenceFilter);

    let count = 0;

    const observer = () => {
        ++count;
    };

    const regId = subject.register(key, observer);
    subject.set(key, true);
    subject.set(key, undefined);

    expect(subject.hasWatchersOrValue(key)).toBeTruthy();

    subject.unregister(key, regId);

    expect(subject.hasWatchersOrValue(key)).toBeFalsy();

    expect(count).toBe(2);
});

test('Set undefined changes nothing when it hasn\'t been set (difference filter)', () => {
    const key = "kiki";
    const subject = new ObservableMap<string, boolean | undefined>(differenceFilter);

    let count = 0;

    const observer = () => {
        ++count;
    };

    subject.register(key, observer);
    subject.set(key, undefined);

    expect(count).toBe(0);
});


test('Set before register, still has value (difference filter)', () => {
    const key = "kiki";
    const subject = new ObservableMap<string, boolean | undefined>(differenceFilter);

    expect(subject.hasWatchersOrValue(key)).toBeFalsy();
    subject.set(key, true);
    expect(subject.hasWatchersOrValue(key)).toBeTruthy();
});


test("Observable counter map (zero or many filter", () => {
    const key = "kiki";
    const subject = new ObservableCounterMap(zeroOrManyFilter);

    let called = 0;

    const callback = () => {
        ++called;
    }

    subject.register(key, callback);

    subject.increment(key);
    subject.increment(key);
    subject.decrement(key);
    subject.decrement(key);

    //Once for the first increment, and one for the last decrement
    expect(called).toBe(2);
});

test("Observable counter map, works properly without observers (zero or many filter)", () => {
    const key = "kiki";
    const subject = new ObservableCounterMap(zeroOrManyFilter);

    subject.increment(key);
    subject.increment(key);

    //Once for the first increment, and one for the last decrement
    expect(subject.hasWatchersOrValue(key)).toBeTruthy();
    expect(subject.get(key)).toBe(2);

    subject.decrement(key);
    subject.decrement(key);

    expect(subject.hasWatchersOrValue(key)).toBeFalsy();
    expect(subject.get(key)).toBeUndefined();
});

test("Zero or many filter", () => {
    expect(zeroOrManyFilter(0, 1)).toBeTruthy();
    expect(zeroOrManyFilter(undefined, 1)).toBeTruthy();

    expect(zeroOrManyFilter(1, 0)).toBeTruthy();
    expect(zeroOrManyFilter(1, undefined)).toBeTruthy();

    expect(zeroOrManyFilter(undefined, 0)).toBeFalsy();
    expect(zeroOrManyFilter(undefined, undefined)).toBeFalsy();
    expect(zeroOrManyFilter(0, undefined)).toBeFalsy();
    expect(zeroOrManyFilter(1, 1)).toBeFalsy();
    expect(zeroOrManyFilter(0, 0)).toBeFalsy();
});

test("Observable counter map (zero or many filter", () => {
    const key = "kiki";
    const subject = new ObservableCounterMap(zeroOrManyFilter);

    let called = 0;

    const callback = () => {
        ++called;
    }

    subject.register(key, callback);

    subject.increment(key);
    subject.increment(key);
    subject.decrement(key);
    subject.decrement(key);

    //Once for the first increment, and one for the last decrement
    expect(called).toBe(2);
});