import { useEffect } from "react";

type TriggerFilter<ValueType> = (newVal: ValueType | undefined, oldVal: ValueType | undefined) => boolean;


class Observable {
    private readonly observers = new Map<number, () => void>();
    private lastId = 0;

    register(callback: () => void) {
        const id = ++this.lastId;
        this.observers.set(id, callback);
        return id;
    }

    unregister(id: number) {
        this.observers.delete(id);
    }

    trigger() {
        for(const [, callback] of this.observers) {
            callback();
        }
    }

    getObserverCount() {
        return this.observers.size;
    }
}


class ObservableValue<ValueType> extends Observable {
    private readonly filter: TriggerFilter<ValueType>;
    private value: ValueType | undefined;

    constructor(filter: TriggerFilter<ValueType | undefined>) {
        super();
        this.filter = filter;
    }

    set(value: ValueType) {
        const shouldTrigger = this.filter(value, this.value);

        this.value = value;

        if(shouldTrigger) {
            this.trigger();
            return true;
        }
        return false;
    }

    get() {
        return this.value;
    }
}

function useWatchObservableValue<T>(cont: ObservableValue<any>, callbacksies: (val: T | undefined) => void) {
    useEffect(() => {
        const sync = () => {
            const val = cont.get();
            callbacksies(val);
        };

        sync();

        const id = cont.register(sync);

        return () => {
            cont.unregister(id);
        }
    }, []);

};

class ObservableCounter extends ObservableValue<number> {
    constructor(filter: TriggerFilter<number | undefined>) {
        super(filter);
    }

    private adjustBy(integer: number) {
        const current = this.get();
        if(!current) {
            this.set(integer);
        } else {
            this.set(current + integer);
        }
    }

    increment() {
        this.adjustBy(1);
    }

    decrement() {
        this.adjustBy(-1);
    }
}


class ObservableMap<KeyType, ValueType> {
    private readonly maps = new Map<KeyType, ObservableValue<ValueType>>();
    private readonly filter: TriggerFilter<ValueType | undefined>;

    private deleteIfYeah(key: KeyType, set: ObservableValue<ValueType>) {
        if(set.getObserverCount() == 0 && set.get() == undefined) {
            //no observers and nothing of note stored, so delete
            this.maps.delete(key);
        }
    }

    private getSetReady(key: KeyType) {
        let go = this.maps.get(key);

        if(!go) {
            go = new ObservableValue<ValueType>(this.filter);
            this.maps.set(key, go);
        }

        return go;
    }

    constructor(filter: TriggerFilter<ValueType | undefined>) {
        this.filter = filter;
    }

    hasWatchersOrValue(key: KeyType) {
        return this.maps.has(key);
    }

    hasCount(key: KeyType) {
        const subMap = this.maps.get(key);

        if(!subMap) {
            return false;
        }

        const result = subMap.get();

        if(!result) {
            return false;
        }

        return true;
    }


    size() {
        return this.maps.size;
    }


    set(key: KeyType, value: ValueType) {
        const set = this.getSetReady(key);
        //const oldVal = set.get();

        const result = set.set(value); //Set handles filtered triggering

        this.deleteIfYeah(key, set);

        return result;
    }

    get(key: KeyType) : ValueType | undefined {
        const set = this.maps.get(key);

        if(set) {
            return set.get();
        }

        return undefined;
    }

    register(key: KeyType, callback: () => void) {
        const set = this.getSetReady(key);
        return set.register(callback);
    }

    unregister(key: KeyType, id: number) {
        let set = this.maps.get(key);

        if(!set) {
            console.warn("Unregistering observer of non-existent set");
            return;
        }

        set.unregister(id);
        this.deleteIfYeah(key, set);
    }

    trigger() {
        for(const [k, v] of this.maps) {
            v.trigger();
        }
    }

    forEach(callback: (k: KeyType, v: ValueType | undefined) => boolean) {
        for(const [k, v] of this.maps) {
            if(!callback(k, v.get()))
                break;
        }
    }
}

class ObservableCounterMap<KeyType> extends ObservableMap<KeyType, number | undefined> {

    constructor(filter: TriggerFilter<number | undefined>) {
        super(filter);
    }

    increment(key: KeyType) {
        let val = this.get(key);
        if(!val) {
            val = 0;
        }

        ++val;
        this.set(key, val);
    }

    decrement(key: KeyType) {
        let val = this.get(key);
        if(!val) {
            val = 0;
        }

        --val;

        if(val == 0) {
            val = undefined;
        }

        this.set(key, val);
    }
}

function differenceFilter(newVal: any, oldVal: any) {
    return newVal != oldVal;
}

function zeroOrManyFilter(newVal: number | undefined, oldVal: number | undefined) : boolean {
    if(!oldVal && !newVal) {
        return false;
    }

    if(oldVal && newVal) {
        if(newVal >= 0 && oldVal >= 0) {
            return false;
        }
    }

    return true;
}

function alwaysTriggerFilter(newVal: any, oldVal: any) {
    return true;
}
function neverTriggerFilter(newVal: any, oldVal: any) {
    return false;
}


function truthyOrNotFilter(newVal: boolean | undefined, oldVal: boolean | undefined): boolean {
    if(newVal && oldVal) return false;
    if(!newVal && oldVal) return true;
    if(newVal && !oldVal) return true;
    if(!newVal && !oldVal) return false;

    return true; //idk man, it's just to shut up typescript the above should cover it...
}

interface UseObserverInterface {
    register: (callback: () => void) => number;
    unregister: (i: number) => void;
}

function useObservable<ContName extends UseObserverInterface>(obsCont: ContName, callback: (cont: ContName) => void) {
    useEffect(() => {
        const obsId = obsCont.register(() => {
            callback(obsCont);
        });

        return () => {
            obsCont.unregister(obsId);
        };
    }, []);
}



export { ObservableValue, 
    ObservableCounter,
    ObservableMap, 
    ObservableCounterMap, 
    Observable, 

    zeroOrManyFilter, 
    alwaysTriggerFilter, 
    neverTriggerFilter, 
    truthyOrNotFilter, 
    differenceFilter,

    useWatchObservableValue,
};