import process from 'node:process';
import nodePath from 'node:path';
import fs from 'node:fs';
import fsPromises from 'node:fs/promises';
import { StringDecoder } from 'node:string_decoder';
import { blake2b } from '@noble/hashes/blake2b';
import { bytesToHex } from '@noble/hashes/utils';
import { WASMagic } from "wasmagic";

const companyFolderName = "Enesda";
const applicationFolderName = "Binder";

/*
function getConfigFilePath(name) {
    return path.join(getConfigDirPath(), name);
}

function getConfigPayload(name) {
    return fs.readFileSync(getConfigFilePath(name), 'utf8');
}

function writeConfigPayload(name, payload) {
    fs.writeFileSync(getConfigFilePath(name), payload, {encoding: "utf8"});
}
*/

let magic;

class FileMonitor {
    #watching = {};
    #webContents = {};

    constructor(webCont) {
        this.#webContents = webCont;
    }

    subscribe(path) {
        if(path in this.#watching) {
            console.warn("Already watching file");
            return;
        }
        try {
            const mon = { path: path, abortSig: new AbortController() };

            fs.watch(path, {persistent: false, recursive: false, signal: mon.abortSig}, (eventType, filename) => {
                this.#webContents.send('fileChange', filename, eventType);
            });
        } catch(e) {
            return;
        }
    }

    unsubscribe(path) {
        if(!(path in this.#watching)) {
            console.warn("Path not in watching during unsubscribe call");
            return;
        }

        this.#watching[path].abortSig.abort();
        delete this.#watching[path];
    }
}

function getSystemConfigDirPath() {
    const systemConfigDir = (() => {
        const windowsAppData = process.env.LOCALAPPDATA;
        if(windowsAppData) {
            return windowsAppData;
        }

        const nixHomeDir = process.env.HOME;

        if(nixHomeDir) {
            return nodePath.join(nixHomeDir, ".config");
        }

        throw new Error("Could not find home directory for config");
    })();

    const companyFolder = nodePath.join(systemConfigDir, companyFolderName);
    const appFolder = nodePath.join(companyFolder, applicationFolderName);

    if (!fs.existsSync(appFolder))
        fs.mkdirSync(appFolder, { recursive: true });

    return appFolder;
}



async function checkExists(path) {
    try {
        await fsPromises.stat(path);
        return true;
    } catch {
        return false;
    }
}


function joinPath(base, second) {
    return nodePath.join(base, second);
}

function baseNameOfPath(path) {
    return nodePath.basename(path);
}

async function writeFile(path, data, encoding) {
    return fsPromises.writeFile(path, data, {encoding: encoding});
}

async function readFile(path, encoding) {
    let result = new Uint8Array();
    let fileHandle;
    let bytesRead = 0;

    try {
        fileHandle = await fsPromises.open(path, 'r');

        while(true) {
            const readResult = await fileHandle.read({position: bytesRead});
            bytesRead += readResult.bytesRead;

            if(readResult.bytesRead) {
                const merged = new Uint8Array(bytesRead);
                merged.set(result);
                const apendee = readResult.buffer.subarray(0, readResult.bytesRead);
                merged.set(apendee, result.length);
                result = merged;
            } else {
                break;
            }
        }
    } catch (e) {
        console.log("Errored out on read: " + e);
    } finally {
        await fileHandle?.close();
    }

    //console.log("read: " + bytesRead + " bytes of: " + path + ", and got: " + result);

    return {data: result, bytesRead: bytesRead};
}

async function appendUtf8file(path, data) {
    return fsPromises.appendFile(path, data, {encoding: "utf8"});
}

async function readUtf8File(path, offset) {
    const decoder = new StringDecoder('utf8');
    let result = "";
    let fileHandle;
    let bytesRead = 0;

    try {
        fileHandle = await fsPromises.open(path, 'r');

        while(true) {
            const readResult = await fileHandle.read({position: offset + bytesRead});
            bytesRead += readResult.bytesRead;

            if(readResult.bytesRead) {
                result += decoder.write(readResult.buffer.subarray(0, readResult.bytesRead));
            } else {
                break;
            }
        }
    } catch (e) {
        console.log("Errored out on read: " + e);
    } finally {
        await fileHandle?.close();
    }

    result += decoder.end();
    //console.log("read: " + bytesRead + " bytes of: " + path + ", and got: " + result);

    return {data: result, bytesRead: bytesRead};
}

async function readMime(path) {
    let fileHandle;
    try {
        fileHandle = await fsPromises.open(path, 'r');

        //thas a lotta bytes.
        //sure, could still be a non-text file even if all of those are valid, but... odds are
        // its not... and we don't know otherwise *anyway* soooo
        const maxBytes = 512;

        const readResult = await fileHandle.read({ length: maxBytes });

        if (!readResult.bytesRead) {
            return undefined;
        }

        if(!magic) {
            magic = await WASMagic.create();

            if(!magic) {
                throw new Error("could not initialize the magic :<");
            }
        }

        return magic.getMime(readResult.buffer.subarray(0, readResult.bytesRead));

    } catch (e) {
        console.log("Errored out on reading, attempting to acquire mime type: " + e);
    } finally {
        await fileHandle?.close();
    }

    return undefined;
}

async function createDirectories(path) {
    //I just want it to always return undefined, fsPromises has a weird conditional return
    // with recursive set to true
    return new Promise(async (resolve, reject) => {
        try {
            await fsPromises.mkdir(path, { recursive: true });
            resolve();
        } catch(e) {
            reject(e);
        }
    });
}

async function readFileModTime(path) {
    return new Promise(async (resolve, reject) => {
        try {
            const stats = await fsPromises.stat(path, {bigint: true});
            resolve(stats.mtimeNs);
        } catch(e) {
            reject();
        }
    });
}

async function readFileStats(path) {
    return new Promise(async (resolve, reject) => {
        try {
            const stats = await fsPromises.stat(path, {bigint: true});
            resolve(stats);
        } catch(e) {
            reject();
        }
    });
}


async function listPaths(path) {
    return new Promise(async (resolve, reject) => {
        let dir;
        try {
            const result = { dirs: [], files: [] };
            dir = await fsPromises.opendir(path);

            for await (const dirent of dir) {
                //Not sure if the isSymlink can be true and the others like isDirectory and isFile
                // still be useful...  I'm hoping so, it's the only sane way to do it
                if (dirent.isDirectory()) {
                    result.dirs.push(dirent.name);
                } else if (dirent.isFile()) {
                    result.files.push(dirent.name);
                }
            }

            resolve(result);
        } catch (e) {
            console.warn("Failed listing paths: " + e);
            dir?.close();
            reject();
        } finally {
            //no it automatically closes for some bloody reason.
            //"wE wAnNa Be NoIsY aBoUt ThIs" but let's not bother being consistent....
        }
    });
}

async function copyPath(from, to) {
    await fsPromises.copyFile(from, to);
}

async function renamePath(path, newPath, maxRetries = 10, retryDelay = 100, retried = 0) {
    if(retried == maxRetries) {
        throw new Error("Could not rename path after max retries");
    }

    try {
        await fsPromises.rename(path, newPath);
        return;
    } catch {
        ++retried;
        const thisRetryDelay = 100 * retried;
        //console.log("Retrying rename");
        return new Promise((resolve, reject) => {
            setTimeout(async () => {
                try {
                    await renamePath(path, newPath, maxRetries, retryDelay, retried);
                    resolve();
                } catch (e) {
                    reject(e);
                }

            }, thisRetryDelay);
        });
    }
}

async function removePath(path) {
    await fsPromises.rm(path, {maxRetries: 10, retryDelay: 100, recursive: true});
}

async function takeChecksumOfFile(path, sumType, maxRetries = 10, timeoutDelay = 0, retried = 0) {


    const digester = (() => {
        if(sumType === "blake2b") {
            return blake2b.create({dkLen: 64}); //oh my fucking gooooooooooooooooooooooooooooooooo
        }

        throw new Error("Could not find checksum digester: " + sumType);
    })();

    //oh are you fucking kidding me, how does this other one even work????
    //it never declares the variable!!!
    let fileHandle;
    let result;

    try {
        fileHandle = await fsPromises.open(path, 'r');

        while(true) {
            const readResult = await fileHandle.read();

            if(!readResult.bytesRead) {
                break;
            }

            digester.update(readResult.buffer.slice(0, readResult.bytesRead));
        }

        result = bytesToHex(digester.digest()).toUpperCase();
    } catch (e) {
        if (retried === maxRetries) {
            console.log("Could not take checksum of file after reaching maximum retries: " + path);
            throw new Error("Max retries reached for taking checksum for reason: " + e);
        }

        ++retried;
        const thisRetryDelay = 100 * retried;
        console.log("File probably busy, trying again in " + thisRetryDelay);
        return new Promise((resolve, reject) => {
            setTimeout(async () => {
                try {
                    const result = await takeChecksumOfFile(path, sumType, maxRetries, timeoutDelay, retried);
                    resolve(result);
                } catch (e) {
                    reject(e);
                }

            }, thisRetryDelay);
        });
    } finally {
        fileHandle?.close();
    }

    return result;
}

export {
    FileMonitor,
    getSystemConfigDirPath,
    checkExists,
    joinPath,
    baseNameOfPath,
    writeFile,
    readFile,
    appendUtf8file,
    readUtf8File,
    readMime,
    createDirectories,
    readFileModTime,
    readFileStats,
    listPaths,
    copyPath,
    renamePath,
    removePath,
    takeChecksumOfFile,
}