import {useState, useEffect} from 'react'


export const NavBarStatorKey = "f";

export class StatorObjectBase
{
    private updateGeneration = 0;
    private listeners = new Set<(i: number) => void>();
    private appContext;

    constructor(appContext: StatorContext)
    {
        this.appContext = appContext;
    }

    getContext()
    {
        return this.appContext;
    }

    registerCallback(callback: (i: number) => void)
    {
        this.listeners.add(callback);
    }

    unregisterCallback(callback: (i: number) => void)
    {
        if(!this.listeners.has(callback))
            throw "CAN'T UNREGISTER A NON-EXISTANT LISTENER";

        this.listeners.delete(callback);
    }


    protected publishChanges()
    {
        ++this.updateGeneration;
        //console.log("Publishing changes to " + this.listeners.size + " listeners")
        for(const callback of this.listeners)
            callback(this.updateGeneration);
    }

    getUpdateGeneration()
    {
        return this.updateGeneration;
    }
}

//I wonder if tagobjectreleationship mapper could be broken up into
// different pieces, like tagview and objectlist, etc? And then
// just watch the necessary stator object...

export class StatorContext
{
    private readonly parent: StatorContext | undefined = undefined;
    private store = new Map<string, StatorObjectBase>();

    constructor(parent: StatorContext | undefined = undefined)
    {
        this.parent = parent;
    }

    private getStatorIfExistsLegacy(statorKey: string): StatorObjectBase | undefined
    {
        const myStator = this.store.get(statorKey);

        if(!myStator)
        {
            if(this.parent)
                return this.parent.getStatorIfExistsLegacy(statorKey);
            else
                return undefined;
        }

        return myStator;
    }

    private getStatorIfExists<T extends StatorObjectBase>(statorClass: new (cont: StatorContext) => T): T | undefined
    {
        const statorKey = statorClass.name;
        const myStator = this.store.get(statorKey);

        if(!myStator)
        {
            if(this.parent)
                return this.parent.getStatorIfExists(statorClass);
            else
                return undefined;
        }

        return myStator as T;
    }

    getStatorLegacy(statorKey: string, initial: StatorObjectBase): StatorObjectBase
    {
        let stator = this.getStatorIfExistsLegacy(statorKey);

        if(stator == undefined)
        {
            this.store.set(statorKey, initial);
            return initial;
        }

        return stator;
    }

    //Construct a new stator if it doesn't exist
    getStator<T extends StatorObjectBase>(statorClass: new (cont: StatorContext) => T): T {
        const statorKey = statorClass.name

        let stator = this.getStatorIfExistsLegacy(statorKey);

        if(stator == undefined) {
            stator = new statorClass(this);
            this.store.set(statorKey, stator);
        }

        return stator as T; //shhhhhhhhhhhhh trust meeee
    }

    useWatchStator<T extends StatorObjectBase>(statorClass: new (cont: StatorContext) => T): T {
        const stator = this.getStator(statorClass);

        const [bodge, setBodge] = useState<number>(stator.getUpdateGeneration());


        useEffect(() =>
        {
            if (stator == undefined)
                return;// dead reference we're done here

            //console.log("register watcher for: " + statorKey)

            const refreshCallback = (i: number) =>
            {
                //console.log("Bodge update: " + i);
                setBodge(i + 1);
            }

            stator.registerCallback(refreshCallback);

            return function()
            {
                //console.log("Unregister watcher of: " + statorKey);
                if (stator == undefined)
                    return;// dead reference

                stator.unregisterCallback(refreshCallback);
            }
        }, [this]); //jesus that's fiendish

        return stator;
    }

    useWatchStatorLegacy<T extends StatorObjectBase>(statorKey: string, initial: T): T
    {
        const stator = this.getStatorLegacy(statorKey, initial);

        const [bodge, setBodge] = useState<number>(stator.getUpdateGeneration());


        useEffect(() =>
        {
            if (stator == undefined)
                return;// dead reference we're done here

            //console.log("register watcher for: " + statorKey)

            const refreshCallback = (i: number) =>
            {
                //console.log("Bodge update: " + i);
                setBodge(i + 1);
            }

            stator.registerCallback(refreshCallback);

            return function()
            {
                //console.log("Unregister watcher of: " + statorKey);
                if (stator == undefined)
                    return;// dead reference

                stator.unregisterCallback(refreshCallback);
            }
        }, [this]); //jesus that's fiendish

        return stator as T;
    }


}

function useStatorContext(parentContext: StatorContext | undefined = undefined)
{
    //kinda ugly
    // I also hate that it creates a new context each render that gets ignored...
    //But it's probably pretty cheap...
    const [statorContext, setStatorContext] = useState<StatorContext>(new StatorContext(parentContext));
    return statorContext;
}


export { useStatorContext }