// @ts-ignore
import Nucleus2, {constructCache} from "./contrib/nucleus2";
import {BinderTag, BinderObject, TagId, ObjectId, Position, BinderConnectionId} from "./backend";
import {v4, validate as validateUUID} from 'uuid'
import { localNodeConfig } from "./localnodeconfig";
import { getMetaApi } from "./ipcbindings";
import { genFracturedPath, openLatestLogFile, openLogFileForIndex } from "./logfile";
import { FileWithPath } from "./filetree";
import { parseTagName } from "./tagnamefunction";


const massOperationFlushLimit = 1000;

const reserved = {
    tagIds: {
        root: "root",
        filenamesParent: "filenames",
        fileModTimestampsParent: "fileModTimestamps",
        objectAddedTimestampParent: "objectAddedTimestamps",
    },
    tagNames: {
        newTag: "New Tag",
        filenames: "Filenames",
        fileModTimesamps: "File Modification Dates",
        objectAddedTimestamp: "Date Added",
    }
}

const systemProtectedTagIds = (() => {
    const result = new Set<string>();

    for(const key in reserved.tagIds) {
        const tagId = reserved.tagIds[key as keyof typeof reserved.tagIds];
        result.add(tagId);
    }

    return result;
})();

const systemLimitChildTagsToSelected = new Set([
    reserved.tagIds.filenamesParent,
    reserved.tagIds.fileModTimestampsParent,
    reserved.tagIds.objectAddedTimestampParent,
]);

const systemProtectedChildRelationships = systemLimitChildTagsToSelected;


const tagsParentTable = "tagParents";
const tagChildrenTable = "tagChildren";
const tagNameTable = "tagNames";
const tagIdToObjectId = "tagIdToObjectId";
const objIdToTagId = "objIdToTagId";
const objectIdToB2Sum = "objectToB2Bsum";
const b2bSumToObjectId = "b2bSumToObjectId";
const tagsTable = "tags";
const objectsTable = "objects";
const mimesTableV1 = "mimesV1";
//For historical reasons, the above are stored at the root but future stuff
// should be stored in the object properties table
const objectPropertiesV1 = "objectPropertiesV1";
const objectPropertyImageWidth = "imageWidth";
const objectPropertyImageHeight = "imageHeight";

const tagChildSortListTable = "tagChildSortList";

const objectMime = "objectMime";
//So, I want this to be multiple choice, but...
// for legacy reasons it is not. But alternative choices
// can go into another one as minors, and this can just
// be considered the preferred
const objectOnDiskPath = "objectOnDiskPath";
const thumbnailOnDiskPath = "thumbnailOnDiskPath";

class TagDatabase {
    private clusterPath: string;
    private engine: Nucleus2;

    //This is which log file we're on. The writelog filename is
    // blake2b(nodeId + writeLogIndex)
    //Because we need to be able to store it nested and we already have a "fractured path"
    // system for storing objects this way, it's just more consistent than some "do 10 per folder or summin idk"
    // approach.
    //And the reason for separate write logs is! Data integrity. So, we don't want to write the whole list
    // *every* time we write a single change as that could quickly get slow. But we don't want to append
    // because that's not atomic and could corrupt to log... We also don't want a new file for each write
    // soooo next best thing is a series of 0.5-1mb-ish files that can be read in for writing modification
    // quickly enough, and also be written quickly enough to a .temp and then renamed for atomicity :3
    //It also has the knock-on effect of making the writelog list simpler to write and parse, as before
    // the idea was to have a comma separated list of json objects... Which needed to be massaged a little
    // to properly parse. But here we can just write a proper json array and be done with it
    //And we can just use the lastRead count to mean record count (which works well for telling us
    // which writeLog file we should be in if we have a specific count per log file)
    private logFileIndex = 0;

    

    private lastWrite = 1; //this way it won't write until lastDirty is marked once.
    private lastDirty = 0;

    private idGen() {
        return v4();
    }

    /*
    private markDirty() {
        this.lastDirty = Date.now();
    }

    flushIfReady() {
        if(this.engine.checkCommitLock() || this.inTransaction) {
            return;
        }

        const saveInterval = 1000; //worst case like 2 seconds, right?
        const currentTime = Date.now();

        if(this.lastWrite <= this.lastDirty && currentTime - this.lastDirty > saveInterval) {
            console.log("Writing to persistent on interval");
            this.engine.flushToPersistent();
            this.lastWrite = Date.now(); //not technically when it *wrote* but yeah
        }
    }
    */


    constructor(path: BinderConnectionId, doneBackCallback: () => void) {
        //There is a BUG TODO:
        //where if the writelog fails on a new tag creation the tag is ghostly existing in the taglist fix it FIXME


        this.clusterPath = path;
        //TODO:::::!!!!! if it's empty!!! add a new tag called "First Tag!" and SELECT IT



        const logReader = async (lastReadByMachine: any) => {
            console.log("Log reading");

            if(!await localNodeConfig.checkBinderExists(path)) {
                console.warn("Binder does not exist, skipping read...");
                return {};
            }

            const writeLogsPath = await localNodeConfig.getWriteLogDirectory(this.clusterPath);
            //wouldn't it be better to have it just store the "lastRead" as the logfile index instead of record count?
            //Sure, we'll end up reading same record multiple times... But if these files are relatively smol..
            // then it should still be faster than opening up several logfiles looking for the one with the right
            // range, right? And much simpler to boot...
            //Maybe a good argument for making them smoler tho... But 1mb worth of records is still probably < 10k records on
            // average...

            const result: any = {};

            //console.log("machine structure: " + JSON.stringify(lastReadByMachine));

            try {
                //This function errors out lol
                const paths = await getMetaApi().listPaths(writeLogsPath);

                if(!paths) {
                    return;
                }

                for (const machineId of paths.dirs) {
                    console.log("doing machine: " + machineId);
                    if (!validateUUID(machineId))
                        continue; //probably a lock file

                    //console.log("for machine: " + machineId);
                    const machineWriteLogPath = await getMetaApi().joinPath(writeLogsPath, machineId);
                    //console.log("machineWriteLogPath: " + machineWriteLogPath);

                    let index = lastReadByMachine[machineId];

                    if(!index) {
                        index = 0;
                    }

                    const machineResult: any = {};
                    machineResult.logs = [];
                    machineResult.readTo = 0;
    
                    while(true) {
                        const logFile = await openLogFileForIndex(machineWriteLogPath, machineId, index);
                        await logFile.load(); //basically no-op if the file doesn't open
    
                        for(const rec of logFile.getRecords()) {
                            //console.log(rec);
                            //const isString = typeof rec == 'string';
                            //console.log("Type of string? " + isString);
                            machineResult.logs.push(JSON.parse(rec)); //nucleus2 has an inconsistent api here...
                        }
    
                        if (logFile.isFull()) { //then there's another file or should be...
                            ++index;
                            continue;
                        }

                        break;
                    }
    
                    machineResult.readTo = index;
                    result[machineId] = machineResult;
                }

                console.log("read all logs for binder @ " + this.clusterPath);

            } catch (e) {
                console.log("Error listing paths: " + e);
            }

            //console.log("Result: " + JSON.stringify(result));
            return result;
        }

        const logWriter = async (logs: Array<string>) => {
            //console.log("writing logs...");

            if(!await localNodeConfig.checkBinderExists(path)) {
                console.warn("Binder does not exist, skipping write...");
                return;
            }

            const nodeId = localNodeConfig.getLocalNodeId();
            //I just spent like 10 minutes trying to figure out where the exception was being thrown
            // and it's not even the right function... It's time to stop...

            const localWriteLogPath = await localNodeConfig.getPathForLocalWriteLogs(this.clusterPath);
            //console.log("Local writelog path: " + localWriteLogPath);
            await getMetaApi().acquireLockfile(localWriteLogPath);

            try {
                const logFile = await openLatestLogFile(localWriteLogPath, nodeId, this.logFileIndex);
                await logFile.load();

                this.logFileIndex = logFile.getIndex(); //sync it cause it could be higher or lower at this point


                //assume not full when opened, cause we check to see if it's full after and then open a new file if so, so
                // it shouldn't ever be full on opening

                logFile.digest(logs);
                await logFile.write(); //somehow forgot this lol

                if (logFile.isFull()) {
                    const nextLogFile = await openLogFileForIndex(localWriteLogPath, nodeId, ++this.logFileIndex);
                    nextLogFile.setPreceding(logFile.getPreceding() + logFile.getCount());
                    await nextLogFile.write(); // force it to exist
                }
            } catch (e) {
                console.log("error :<" + e);
            } finally {
                //console.log("Releasing");
                await getMetaApi().releaseLockfile(localWriteLogPath);
                //console.log("Done release");
            }
        }

        const cacheWriter = async (cachey: string) => {

        };

        this.engine = new Nucleus2(localNodeConfig.getLocalNodeId(), logReader, logWriter, constructCache, cacheWriter);

        (async () => {
            await this.engine.sync();



            doneBackCallback();
        })();
    }

    private deprecatedRecursivelyPopulateBinderTag(parent: TagId, result: Array<BinderTag>) {
        const childIds = this.getSortedTags(parent);

        for(const id of childIds) {

            //First call of this function is with root, which cannot be deleted, so we
            // only process children of root, and then it's implied that they've been checked
            // before the recurse

            const tagGetTestVal =  this.engine.get(tagsTable, id);
            const name = this.engine.get(tagNameTable, id);

            if(!tagGetTestVal) {
                console.warn("Ghost tag detected (failed delete?), attempting re-delete");
               // console.log("Ghost tag: " + id + ", " + tagGetTestVal + ", name: " + name);
                this.deleteTag(id);
                continue;
            }

            const intermediate = new BinderTag();
            intermediate.id = id;
            intermediate.name = name;

            this.deprecatedRecursivelyPopulateBinderTag(id, intermediate.children);

            result.push(intermediate);
        }
    }

    deprecratedAllTagsEndpoint(): Array<BinderTag> {
        const result = new Array<BinderTag>();
        this.deprecatedRecursivelyPopulateBinderTag(reserved.tagIds.root, result);
        return result;
    }

    getTagParentChain(tagId: TagId): Array<TagId> {
        const result = new Array<TagId>();
        if(tagId == reserved.tagIds.root) {
            return result;
        }

        let working = tagId;
        while(true) {
            working = this.engine.get(tagsParentTable, working);

            if(working == reserved.tagIds.root) {
                break;
            }

            if(!working) {
                console.warn("Tag database weird, tag apparently doesn't have parent (no root)");
                break;
            }

            result.push(working);
        }

        return result;
    }

    getTagParent(tagId: TagId) {
        if(!tagId || tagId == reserved.tagIds.root) {
            return undefined;
        }
        return this.engine.get(tagsParentTable, tagId);
    }

    private loadObject(objId: ObjectId): BinderObject {
        const result = new BinderObject();

        result.id = objId;
        result.b2bsum = this.engine.get(objectIdToB2Sum, objId);
        //.index is built by tagobjectrelationshipmapper... for some reason
        //so no bother with that.
        result.mime = this.engine.get(objectMime, objId);

        result.objectOnDiskPath = this.engine.get(objectOnDiskPath, objId);
        result.thumbnailOnDiskPath = this.engine.get(thumbnailOnDiskPath, objId);

        result.tags = new Set(...this.engine.listMinors(objIdToTagId, objId));

        return result;
    }

    private recursivelyGetObjectsInclusive(tagId: TagId, result: Set<ObjectId>) {
        if(this.checkLimitTagChildrenToObjectSelection(tagId)) {
            return;
        }

        const myChildren = this.getObjectIdsWithTag(tagId);

        for(const myChildObject of myChildren) {
            result.add(myChildObject);
        }

        const sortedChildren = this.getSortedTags(tagId); //oh thank god this is sync.

        for(const childTag of sortedChildren) {
            this.recursivelyGetObjectsInclusive(childTag, result);
        }
    }

    getObjectsWithTagInclusive(tagId: TagId) {
        const result = new Set<ObjectId>();
        this.recursivelyGetObjectsInclusive(tagId, result);
        return result;
    }

    getObjectIdsWithTag(tagId: TagId) {
        return this.engine.listMinors(tagIdToObjectId, tagId);
    }

    getTagIdsOnObject(objId: ObjectId) {
        return this.engine.listMinors(objIdToTagId, objId);
    }

    objectsForTagEndpoint(tags: Array<TagId>): Array<BinderObject> {
        const visitedObjectIds = new Array<string>();
        const result = new Array<BinderObject>();


        for(const tagId of tags) {
            const objectIds = this.engine.listMinors(tagIdToObjectId, tagId);

            //console.log("Minors: " + JSON.stringify(objectIds));

            for(const objId of objectIds) {
                //console.log("Listing minor: " + objId);
                if(visitedObjectIds.includes(objId)) {
                    continue;
                }

                result.push(this.loadObject(objId));
                visitedObjectIds.push(objId);
            }
        }

        return result;
    }

    keywordSeachTags(query: string) {
        const result = new Map<string, Set<TagId>>();
        const namedTags = this.engine.listMinors(tagNameTable);

        const substrings = query.split(" ");


        for(const tagId of namedTags) {
            const name = parseTagName(this.getTagName(tagId));

            for (const sub of substrings) {
                let subSet = result.get(sub);
                if(!subSet) {
                    subSet = new Set<TagId>();
                    result.set(sub, subSet);
                }
                //We must be in fucking topsy turvy land because this
                // is only true if it isn't a match? Or are we dealing we type
                // coercion...
                if (name.toLocaleLowerCase().includes(sub.toLocaleLowerCase())) {
                    subSet.add(tagId);
                }
            }
        }

        return result;
    }


    private addTagToObjectInternal(theTag: TagId, objId: ObjectId) {
        /*
        if(this.checkLockedRelationships(theTag)) {
            throw new Error("Cannot modify relationship of: " + theTag);
        }
        */
        this.engine.set(tagIdToObjectId, theTag, objId, true);
        this.engine.set(objIdToTagId, objId, theTag, true);
    }

    private removeTagFromObjectInternal(theTag: TagId, objId: ObjectId) {
        /*
        if(this.checkLockedRelationships(theTag)) {
            throw new Error("Cannot modify relationship of: " + theTag);
        }
        */
        this.engine.del(tagIdToObjectId, theTag, objId);
        this.engine.del(objIdToTagId, objId, theTag);
    }


    async addTagToObjects(theTag: TagId, theObjects: Array<ObjectId>) {
        let c = 0;
        for (const objId of theObjects) {
            if(c++ == massOperationFlushLimit) {//lol c++
                c = 0;
                await this.engine.flushToPersistent();
            }

            //The order here is arbitrary but important, check
            // registerObjectTagLinkChange for details
            this.addTagToObjectInternal(theTag, objId);
        }

        if(c) {
            await this.engine.flushToPersistent();
        }
    }


    async removeTagFromObjects(theTag: TagId, theObjects: Array<ObjectId>) {
        let count = 0;
        for(const objId of theObjects) {
            if(count++ == massOperationFlushLimit) {
                count = 0;
                await this.engine.flushToPersistent();
            }
            //The order here is arbitrary but important, check
            // registerObjectTagLinkChange for details
            this.removeTagFromObjectInternal(theTag, objId);
        }

        if(count) { //otherwise it flushed on the last operation
            await this.engine.flushToPersistent();
        }
    }

    private createTagInternal(theName: string, tagParent: TagId, tagId: TagId) {
        this.engine.set(tagsTable, tagId, true);
        const properName = theName ? theName : "New Tag";
        this.engine.set(tagNameTable, tagId, properName);
        let chosenParent = reserved.tagIds.root;
        if(tagParent !== "")
            chosenParent = tagParent;

        this.engine.set(tagChildrenTable, chosenParent, tagId, true);
        this.engine.set(tagsParentTable, tagId, chosenParent);


        /*
        const result = new BinderTag();
        result.name = theName;
        result.id = tagId;
        //no children cause it's just a newborn alright? jeez

         */
    }

    checkLimitTagChildrenToObjectSelection(tag: TagId) {
        //In the future of course, this should be a lookup in a database table to apply it to user tags
        return systemLimitChildTagsToSelected.has(tag);
    }

    checkLockedRelationships(tag: TagId) {
        return systemProtectedChildRelationships.has(this.getTagParent(tag));
    }

    /*
    private bakeTagSortList(tag: TagId) {
        //So when we create a tag or reparent it, or delete a tag, getSortedTags
        // *implicitly* knows what to do, but it doesn't change the actual stored list.
        //Well, if we have two tags added then the order becomes implicit...
        // of course.......
        // now..........
        // what if a change comes over the wire............
        // like a reparent................................

        //okay I'm overthinking this, if it's reparented from over the wire, then it'll ALSO
        // update the tagsort list...
        //buttttttttttt it'll be wrong......

        //Why not just watch both, have a function to check if it needs baking, and if it does, 
        // do that and return?
        const sortedTags = this.getSortedTags(tag);
        this.engine.set(tagChildSortListTable, tag, sortedTags);
        this.markDirty();
    }
    */

    async newTagToObjects(theName: string, tagParent: TagId, objects: Array<ObjectId>) {
        const tagId = await this.createTagEndpoint(theName, tagParent);
        this.addTagToObjects(tagId, objects);
        await this.engine.flushToPersistent();
    }

    private createUserTagInternal(theName: string, tagParent: TagId) {
        /*
        if(systemProtectedChildRelationships.has(tagParent)) {
            throw new Error("Cannot add new tag as child of: " + tagParent);
        }
        */
        const tagId = this.idGen();
        this.createTagInternal(theName, tagParent, tagId);
        return tagId;
    }

    //NOTE! this is slightly incompatible with the old version that just returned
    // the tagId
    async createTagEndpoint(theName: string, tagParent: TagId) {
        const tagId = this.createUserTagInternal(theName, tagParent);
        await this.engine.flushToPersistent();
        return tagId;
    }

    //this one has a legitimate use case, because it's doing repositioning as a child tag...
    // except that that could be done here...... TODO
    createTagWithoutFlush(theName: string, tagParent: TagId) {
        const tagId = this.createUserTagInternal(theName, tagParent);
        return tagId;
    }

    async renameTagEndpoint(tagId: TagId, theNewName: string) {
        if(systemProtectedTagIds.has(tagId)) {
            return false;
        }
        this.engine.set(tagNameTable, tagId, theNewName);
        await this.engine.flushToPersistent();
        return true;
    }


    async setTagParent(tag: TagId, newParentId: TagId) {
        if(this.checkLockedRelationships(tag)) {
            return;
        }

        if(tag == newParentId)
            return false;

        const currentParent = this.engine.get(tagsParentTable, tag);

        if(!currentParent && !newParentId)
            return false; //no-op

        //Block if any parent of the new parent is the new child

        for(let working = newParentId; working !== reserved.tagIds.root; working = this.engine.get(tagsParentTable, working)) {
            if(working == tag) {
                console.log("Skipping reparent because of genealogical paradox");
                return false; //Got you ;)
            }
        }

        //if(optionalCurrentParent) //there's always a parent now, it's called root

        if(currentParent == newParentId)
            return false;

        this.engine.del(tagChildrenTable, currentParent, tag);
        this.engine.del(tagsParentTable, tag); //should just overwrite anyway but w/e

        //set parent FIRST, so that it can be read when the children table major observer fires
        this.engine.set(tagsParentTable, tag, newParentId); 
        this.engine.set(tagChildrenTable, newParentId, tag, true);


        await this.engine.flushToPersistent();

        return true;
    }

    checkTagExists(tag: TagId) {
        return this.engine.get(tagsTable, tag) !== undefined;
    }

    chekcObjectExists(objId: ObjectId) {
        return this.engine.get(objectsTable, objId) !== undefined;
    }

    getTagName(tag: TagId) {
        return this.engine.get(tagNameTable, tag);
    }

    checkTagHasChildren(tag: TagId) {
        return this.engine.listMinors(tagChildrenTable, tag).length != 0;
    }

    getSortedTags(tag: TagId) {
        //We lazily update this list, cause it's more efficient to remove them when we read it
        // than every single time we remove a tag... I guess.

        let masterList = new Set<TagId>(this.engine.listMinors(tagChildrenTable, tag));
        if (!masterList)
            masterList = new Set<TagId>();

        let sortList = this.engine.get(tagChildSortListTable, tag);

        if(!sortList)
            sortList = new Array<TagId>();


        for(const child of masterList) {
            if(!sortList.includes(child)) {
                sortList.splice(0, 0, child); //splice into the front
            }
        }

        const result = new Array<TagId>();

        for(const child of sortList) {
            if(!masterList.has(child))
                continue;

            result.push(child);
        }


        return result;
    }

    repositionTagWithoutFlush(tagToMove: TagId, tagToMoveTo: TagId, pos: Position) {
        if(this.checkLockedRelationships(tagToMove)) {
            return false;
        }

        if(tagToMove == tagToMoveTo) {
            return false;
        }


        if(!this.checkTagExists(tagToMove))
            throw new Error("Client error: could not find tag to move");


        if(!this.checkTagExists(tagToMoveTo))
            throw new Error("Client error: could not find tag to move to");

        //okay so first we have to determine whether they share a parent... Reparent if they do
        // then add to the tag list at a specific position

        let parentProper: TagId = reserved.tagIds.root;

        {
            const optionalParent = this.engine.get(tagsParentTable, tagToMove);
            const optionalOtherParent = this.engine.get(tagsParentTable, tagToMoveTo);

            if(optionalParent != optionalOtherParent)
            {
                if(!optionalOtherParent) {
                    parentProper = reserved.tagIds.root; 
                } else {
                    parentProper = optionalOtherParent;
                }

                if(!this.setTagParent(tagToMove, parentProper)) {
                    return false;
                }
            }
            else if(optionalParent) {
                parentProper = optionalParent;
            }

            //otherwise it's root anyway
        }


        const tagList = this.getSortedTags(parentProper);

        let currentPos = 0;

        let foundTagToMove = false;
        let foundTarget = false;

        while(currentPos != tagList.length) {
            if(tagList[currentPos] == tagToMove) {
                tagList.splice(currentPos, 1);
                foundTagToMove = true;
            } else if(tagList[currentPos] == tagToMoveTo) {
                if(pos == Position.Before) {
                    //BAD AND YOU SHOULD FEEL BADDDDDDD
                    //this isn't javascript c'monnnn
                    //currentPos = tagList.insert(currentPos, tagToMove);

                    tagList.splice(currentPos, 0, tagToMove);

                    //Splice inserts *before* the position given.
                    // and the currentPosition points to the inserted.
                    //We want to skip both values or we'll have a real bad
                    // time in this loop.
                    //We know currentPos wasn't end, and we know now that there
                    // is one element after it. So, we should be safe
                    currentPos += 2;
                } else if(pos == Position.After) {
                    //Insert positions *before* the iterator passed in,
                    // so we have to go past the one we to insert *after*
                    ++currentPos; //This could be end! but it's okay

                    tagList.splice(currentPos, 0, tagToMove);
                    //currentPos = tagList.insert(currentPos, tagToMove);
                    ++currentPos; //We need to move past the one we just inserted
                    //Could be end again!
                } else throw new Error("this can't happen");

                foundTarget = true;
            } else ++currentPos;

            if(foundTarget && foundTagToMove)
                break;
        }

        //Throwing an exception here isn't the greatest for performance, but neither
        // is opening up a copy of the database on disk to write a bunch of junk to

        //throwing out of here is safe because we've so far only mutated a copy of the tag list
        if(!foundTarget)
            throw new Error("Client error: could not find target tag to position around in taglist");

        if(!foundTagToMove)
            throw new Error("Client error: could not find tag to move in taglist");

        this.engine.set(tagChildSortListTable, parentProper, tagList); //:)


        return true;

    }

    async repositionTag(tagToMove: TagId, tagToMoveTo: TagId, pos: Position) {
        if(this.repositionTagWithoutFlush(tagToMove, tagToMoveTo, pos)) {
            await this.engine.flushToPersistent();
        }
    }

    private deleteTagWithoutFlushing(tag: TagId) {
        /*
        if(this.checkLockedRelationships(tag)) {
            return;
        }
        */

        if(systemProtectedTagIds.has(tag)) {
            return;
        }
        //Delete children first that way callbacks can still make sane requests regarding parental information
        //Depth first is just easier to grok from the outside
        const children = this.engine.listMinors(tagChildrenTable, tag);

        for(const childId of children) {
            if(childId == reserved.tagIds.root) {
                console.warn("child tag id is root for some reason???");
                continue;
            }

            if(systemProtectedTagIds.has(childId)) {
                this.setTagParent(childId, reserved.tagIds.root);
            } else {
                //This call deletes the minor entry so we gud
                this.deleteTagWithoutFlushing(childId);
            }
        }

        //Remove tagged objects before deleting the tag
        const objects = this.engine.listMinors(tagIdToObjectId, tag);

        for(const objId of objects) {
            this.engine.del(objIdToTagId, objId, tag);
            this.engine.del(tagIdToObjectId, tag, objId);
        }

        //This mighht be called with any of these tables missing the tag in case of partial sync or failed
        // delete. Should be rare, but probably possible.

        //Delete tag *before* deleting the parental relationship, so that
        // we can still query parental relationships as they were from the callback
        this.engine.del(tagsTable, tag);


        const parentTagId = this.engine.get(tagsParentTable, tag);
        this.engine.del(tagChildrenTable, parentTagId, tag);
        this.engine.del(tagsParentTable, tag);

        this.engine.del(tagNameTable, tag);
    }

    async deleteTag(tag: TagId) {
        this.deleteTagWithoutFlushing(tag);
        await this.engine.flushToPersistent();
    }

    deleteObject(obj: ObjectId) {
        //Delete object, THEN links, because we need the tag list to clean up...

        throw new Error(".deleteObject not implemented in TagDatabase");
    }

    //These are gonna have to co-operate with the backend...

    //Why isn't it just "additionalTags" with an "at least one" requirement?
    createObjectsDeprecated(baseTag: TagId, additionalTags: Array<TagId>, theFiles: Array<string>) {
        //Doing this properly is super complicated but for tonight, the proof of concept,
        // shall be simpo :3

        throw new Error("This function has been deprecated");
    }

    pickFilename(objId: ObjectId) {
        const tags = this.getTagIdsOnObject(objId);

        for(const t of tags) {
            if(this.getTagParent(t) == reserved.tagIds.filenamesParent) {
                //it would be nice if we correlated this to the sorted tags list of reservedfor filenames...
                // that way you could pick it by moving it
                return this.getTagName(t); //just go for the first one it's too hard otherwise...
            }

        }

        return this.engine.get(objectIdToB2Sum, objId);
    }

    private createReservedFileTagsWithoutFlush() {
        let fileStuffTagId: TagId | undefined;

        const getFileStuffTagId = () => {
            if(!fileStuffTagId) {
                fileStuffTagId = this.createTagWithoutFlush("File Stuff", reserved.tagIds.root);

                if(!fileStuffTagId) {
                    throw new Error("Could not create tag for file stuff");
                }
            }
            return fileStuffTagId;
        }

        if (!this.checkTagExists(reserved.tagIds.filenamesParent)) {
            this.createTagInternal(reserved.tagNames.filenames, getFileStuffTagId(), reserved.tagIds.filenamesParent);
        }

        if (!this.checkTagExists(reserved.tagIds.fileModTimestampsParent)) {
            this.createTagInternal(reserved.tagNames.fileModTimesamps, getFileStuffTagId(), reserved.tagIds.fileModTimestampsParent);
        }

        if (!this.checkTagExists(reserved.tagIds.objectAddedTimestampParent)) {
            this.createTagInternal(reserved.tagNames.objectAddedTimestamp, getFileStuffTagId(), reserved.tagIds.objectAddedTimestampParent);
        }
    }

    async uploadFile(file: FileWithPath, inputTags: Array<TagId>) {
        if(!await localNodeConfig.checkBinderExists(this.clusterPath)) {
            console.warn("Skipping upload because binder does not exist");
            return;
        }

        this.createReservedFileTagsWithoutFlush();
        console.log("uploading: " + file.path);
        const stageFolder = await localNodeConfig.getPathForStage(this.clusterPath);
        console.log("taking initial checksum");
        const unstagedChecksum = await getMetaApi().takeChecksumOfFile(file.path, "blake2b");
        //probably shouldn't copy it into the stage until we confirm it's not a duplicate...
        // more efficient... TODO I just want to be conservative with editing this very critical function
        // that is fairly well tested now
        const stagePath = await getMetaApi().joinPath(stageFolder, unstagedChecksum);

        const modificationDate = await getMetaApi().readFileModTime(file.path);
        const dateAddedTimestamp = Date.now() * 1000000;

        const formatTimestampName = (time: any) => {
            return `!!binder-function-time(${time})!!`;
        }

        const modDateTagName = formatTimestampName(modificationDate);
        const addedDateTagName = formatTimestampName(dateAddedTimestamp);

        //!!binder-function-time(1367037119000000000)!!
        await getMetaApi().copyPath(file.path, stagePath);

        const objectsFolder = await localNodeConfig.getPathForObjects(this.clusterPath);

        const fracturedObjectPath = await genFracturedPath(objectsFolder, unstagedChecksum);
        const objectPath = await getMetaApi().joinPath(fracturedObjectPath, unstagedChecksum);

        //then we're pretty safe to start writing to database, we've copied the file (which is not atomic)

        let objectId = this.engine.get(b2bSumToObjectId, unstagedChecksum);
        let filenameTagId: string | undefined;
        let addDateAlreadyExists = false;
        let modDateTagId: string | undefined;

        const doCopy = async () => {
            await getMetaApi().createDirectories(fracturedObjectPath);
            console.log("Double checking checksum")
            const doubleCheckedSum = await getMetaApi().takeChecksumOfFile(stagePath, "blake2b");
            if (doubleCheckedSum != unstagedChecksum) {
                console.log("Checksum: " + doubleCheckedSum);
                console.log("Does not match previous: " + unstagedChecksum);
                throw new Error("Data corruption at the disco! Corrupted during file copy");
            } else { //else just to make damn sure I don't change the above and have this fall thru
                await getMetaApi().renamePath(stagePath, objectPath);
            }
        }

        if (objectId && !await getMetaApi().checkPathExists(objectPath)) { //we have an entry but failed to copy apparently
            console.log("We have a partial, friends!");
            await doCopy();

        } else if (!objectId) {
            //copy first, which double checks the checksum, to make DAMN sure we have the thing
            // otherwise we could not have it but still deduplicate... Or it could be a bad copy
            //Or something...
            await doCopy();

            objectId = this.idGen();
            this.engine.set(objectsTable, objectId);
            this.engine.set(objectIdToB2Sum, objectId, unstagedChecksum);
            this.engine.set(b2bSumToObjectId, unstagedChecksum, objectId);
        } else { //already exists

            //or does it??? check for corruption. We overwrite any file with a checksum that doesn't match it's contents
            const existsAndIntact: boolean = await (async () => {
                const existingObjectPathDirs = await this.getObjectPath(objectId);

                if (!existingObjectPathDirs) {
                    return false;
                }
                if (!existingObjectPathDirs.path) {
                    return false;
                }

                const existingObjectChecksum = await getMetaApi().takeChecksumOfFile(existingObjectPathDirs.path, "blake2b");

                return existingObjectChecksum == unstagedChecksum;
            })();

            //don't worry, we only have a hit if the unstagedChecksum matches a checksum in the database, so
            // it's very unlikely we're replacing something of value with something broken. And we're only
            // overwriting it if the stored checksum doesn't match the stored file

            if (existsAndIntact) {
                await getMetaApi().removePath(stagePath);
                //The idea is that 1) deduplication itself isnt' super common, and that 2)
                // objects will typically have a smol number of tags. And honestly < 100 it's probably
                // faster to do a linear search anyway. So...
                for (const t of this.getTagIdsOnObject(objectId)) {
                    const tagParent = this.getTagParent(t);
                    if (tagParent == reserved.tagIds.filenamesParent) {
                        if (this.getTagName(t) == file.name) {
                            filenameTagId = t;
                        }
                    } else if (tagParent == reserved.tagIds.fileModTimestampsParent) {
                        if (this.getTagName(t) == modDateTagName) {
                            modDateTagId = t;
                        }
                    } else if (tagParent == reserved.tagIds.objectAddedTimestampParent) {
                        addDateAlreadyExists = true;
                    }

                    if (filenameTagId && modDateTagId && addDateAlreadyExists) {
                        break;
                    }
                }
            } else {
                console.warn("overwriting existing file because its checksum doesn't match stored, assuming corrupt");
                await doCopy();
            }

        }

        if (!filenameTagId) {
            filenameTagId = await this.createTagEndpoint(file.name, reserved.tagIds.filenamesParent);
            if (!filenameTagId) {
                throw new Error("Failed to acquire filename tag id"); //which can never happen but it makes typescript happy so w/e
            }
        }

        this.addTagToObjects(filenameTagId, [objectId]);

        if (!modDateTagId) {
            modDateTagId = await this.createTagEndpoint(modDateTagName, reserved.tagIds.fileModTimestampsParent);
            if (!modDateTagId) {
                throw new Error("Failed to acquire mod date tag Id");
            }
        }

        this.addTagToObjects(modDateTagId, [objectId]);

        if (!addDateAlreadyExists) {

            const addDateTagId = await this.createTagEndpoint(addedDateTagName, reserved.tagIds.objectAddedTimestampParent);
            this.addTagToObjects(addDateTagId, [objectId]);
        }

        //But... Modification date collisions are much less uhhh uncommon? And date added..
        // hmm should that be deduped?

        //if... we assume that tags are rarely renamed, then keeping them in their own reverse lookup table
        // shouldn't be orders of magnitude more expensive than not...
        //But does it matter if we deduplicate tagnames in filenames?
        //We could keep a separate filename -> tagid thing but then if the tag gets renamed it'll screw up the referecne
        // (we could special case that and make another...)
        //Or should filenames just be exclusive?
        //The only problem I foresee is if we have a filename from uploading the same folder, same named files, multiple
        // times. Then we just spam the filenames tag with duplicates. What if we simply search for the filename
        // in the object IF it's a duplicate and not otherwise? Shoulnd't be too costly...


        for (const tagId of inputTags) {
            this.addTagToObjects(tagId, [objectId]);
        }

        await this.engine.flushToPersistent();
    }

    getMimeIfExists(objectId: string) {
        return this.engine.get(mimesTableV1, objectId);
    }

    //This is usually a batch operation and could potentially be faster (if it's a bottleneck haven't test)
    // by having just an input of a bunch of objectIds, and we just set them all in one go and flush...
    // although it's async so... ughh...
    //Transactions don't work the way they should...
    async getMimeGenIfNotExists(objectId: string) {
        if(!await localNodeConfig.checkBinderExists(this.clusterPath)) {
            console.warn("Skipping mime check because binder does not exist");
            return undefined;
        }

        let mime = this.getMimeIfExists(objectId);

        if(mime) {
            return mime;
        }

        const checksum = this.engine.get(objectIdToB2Sum, objectId);

        if(!checksum) {
            return undefined;
        }

        const objectsFolder = await localNodeConfig.getPathForObjects(this.clusterPath);
        const fracturedObjectPath = await genFracturedPath(objectsFolder, checksum);
        const objectPath = await getMetaApi().joinPath(fracturedObjectPath, checksum);

        mime = await getMetaApi().readMime(objectPath);

        if(!mime) {
            throw new Error("could not generate mime");
        }

        this.engine.set(mimesTableV1, objectId, mime);

        await this.engine.flushToPersistent();

        return mime;
    }

    async getThumbnailPath(objectId: string) {
        if(!await localNodeConfig.checkBinderExists(this.clusterPath)) {
            console.warn("Skipping get thumbnail path because binder does not exist");
            return {directory: "", path: ""};
        }

        const thumbnailsFolder = await localNodeConfig.getPathForThumbnails(this.clusterPath);
        const checksum = this.engine.get(objectIdToB2Sum, objectId);

        if(!checksum) {
            return {directory: "", path: ""};
        }

        const fracturedThumnailFolder = await genFracturedPath(thumbnailsFolder, checksum);
        const thumbnailPath = await getMetaApi().joinPath(fracturedThumnailFolder, checksum);

        return {directory: fracturedThumnailFolder, path: thumbnailPath};
    }

    async getObjectPath(objectId: string) {
        if(!await localNodeConfig.checkBinderExists(this.clusterPath)) {
            console.warn("Skipping get object path because binder does not exist");
            return "";
        }

        const objectsFolder = await localNodeConfig.getPathForObjects(this.clusterPath);
        const checksum = this.engine.get(objectIdToB2Sum, objectId);

        if(!checksum) {
            return "";
        }

        const fracturedObjectsFolder = await genFracturedPath(objectsFolder, checksum);
        const objectPath = await getMetaApi().joinPath(fracturedObjectsFolder, checksum);

        return {directory: fracturedObjectsFolder, path: objectPath};
    }

    registerTagParentChangeObserver(obs: (parent: string, child: string, rel: boolean, unsub: () => void) => void) {
        return this.engine.registerMajorObserver(tagChildrenTable, (keys: string[], val: any, unsub: () => void) => {
            //keys -> table, parent, child
            //console.log("Tag parent change observer, keys: " + JSON.stringify(keys) + ", val: " + JSON.stringify(val));
            const parent = keys[1];
            const child = keys[2];

            const rel = val ? true : false;

            //Don't check if the tags exist because they might in the future...
            //eventually consistent...

            obs(parent, child, rel, unsub);
        });
    }

    removeTagParentChangeObserver(id: number) {
        this.engine.removeMajorObserver(tagChildrenTable, id);
    }

    //This one is different from the other because it's generic and always fires
    // whereas the other is specific to a tag
    registerTagChildRepositionObs(obs: (parentTagId: TagId, unsub: () => void) => void) {
        return this.engine.registerMajorObserver(tagChildSortListTable, (keys: string[], val: any, unsub: () => void) => {
            //keys -> table, parent, child
            //console.log("Tag parent change observer, keys: " + JSON.stringify(keys) + ", val: " + JSON.stringify(val));
            const parent = keys[1];
            //const list = keys[2];

            obs(parent, unsub);
        });
    }

    removeTagChildRepositionObs(id: number) {
        this.engine.removeMajorObserver(tagChildSortListTable, id);
    }

    registerTagDeleteObserver(obs: (tagId: TagId, unsub: () => void) => void) {
        return this.engine.registerMajorObserver(tagsTable, (keys: string[], val: any, unsub: () => void) => {
            //keys -> table, tagId

            if(val) { //was created, not deleted
                return; 
            }

            const tagId = keys[1];
            obs(tagId, unsub);
        });
    }

    removeTagDeleteObserver(id: number) {
        this.engine.removeMajorObserver(tagsTable, id);
    }

    registerObjectDeleteObserver(obs: (objId: ObjectId, unsub: () => void) => void) {
        return this.engine.registerMajorObserver(objectsTable, (keys: string[], val: any, unsub: () => void) => {
            //keys -> table, objId

            if(val) { //created, not deleted
                return;
            }

            const objId = keys[1];
            obs(objId, unsub);
        });
    }

    removeObjectDeleteObserver(id: number) {
        this.engine.removeMajorObserver(objectsTable, id);
    }

    registerObjectTagLinkChangeObserver(obs: (tagId: TagId, objectId: ObjectId, rel: boolean, unsub: () => void) => void) {
        //We could in theory monitor either tagIdToObjectId, or ObjIdToTagId, but the latter is run second
        // and we want both to exist when we fire the observer.
        return this.engine.registerMajorObserver(objIdToTagId, (keys: string[], val: any, unsub: () => void) => {
            //keys -> table, objId, tagId
            const objId = keys[1];
            const tagId = keys[2];

            const rel = val ? true : false;

            obs(tagId, objId, rel, unsub);
        });
    }

    removeObjectTagLinkChangeObserver(id: number) {
        this.engine.removeMajorObserver(objIdToTagId, id);
    }

    registerTagNameChangeObserver(tagId: TagId, callback: (name: string) => void) {
        return this.engine.registerPropertyObserver(tagNameTable, tagId, callback);
    }

    removeTagNameChangeObserver(tagId: TagId, id: number) {
        this.engine.removePropertyObserver(tagNameTable, tagId, id);
    }

    registerChildrenOrderChange(tagId: TagId, callback: () => void) {
        return this.engine.registerPropertyObserver(tagChildSortListTable, tagId, callback);
    }

    removeChildrenOrderChangeObserver(tagId: TagId, id: number) {
        this.engine.removePropertyObserver(tagChildSortListTable, tagId, id);
    }

    registerSpecificChildrenObserver(tagId: TagId, callback: (added: boolean) => void) {
        return this.engine.registerIndexObserver(tagChildrenTable, tagId, callback);
    }

    removeSpecificChildrenObserver(tagId: TagId, id: number) {
        this.engine.removeIndexObserver(tagChildrenTable, tagId, id);
    }


    exportObject(dest: string, objIds: Array<ObjectId>) {
    }

    async syncCluster() {
        await this.engine.sync();
    }
}


export {TagDatabase,
    reserved,
}